r"""This module defines the base class for the experiment trackers."""

__all__ = [
    "BaseBasicExpTracker",
    "BaseExpTracker",
    "NotActivatedExpTrackerError",
]

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Optional, Union

from objectory import AbstractFactory

from gravitorch.utils.artifacts.base import BaseArtifact
from gravitorch.utils.exp_trackers.steps import Step
from gravitorch.utils.integrations import is_matplotlib_available, is_pillow_available

if is_matplotlib_available():
    from matplotlib.pyplot import Figure
else:
    Figure = "matplotlib.pyplot.Figure"  # pragma: no cover

if is_pillow_available():
    from PIL.Image import Image
else:
    Image = "PIL.Image.Image"  # pragma: no cover


class BaseExpTracker(ABC, metaclass=AbstractFactory):
    r"""Defines the base class to implement an experiment tracker.

    It is recommended to use the experiment tracker in a context
    manager. Using a context manager ensures the ``start`` and ``end``
    methods are called properly. It is possible to manually call the
    ``start`` and ``end`` methods. It is recommended to not mix these
    two approaches to avoid unexpected side-effects.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.exp_trackers import BaseExpTracker
        # Context manager approach (recommended)
        >>> my_exp_tracker: BaseExpTracker = ...  # Initialize an experiment tracker object
        >>> print(my_exp_tracker.is_activated())
        False
        >>> with my_exp_tracker as exp_tracker:
        ...     print(exp_tracker.is_activated())
        True
        >>> print(my_exp_tracker.is_activated())
        False

        # Manual approach
        >>> exp_tracker: BaseExpTracker = ...  # Initialize an experiment tracker object
        >>> print(exp_tracker.is_activated())
        False
        >>> exp_tracker.start()
        >>> print(exp_tracker.is_activated())
        True
        >>> exp_tracker.end()
        >>> print(exp_tracker.is_activated())
        False
    """

    def __enter__(self) -> "BaseExpTracker":
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end()

    @property
    @abstractmethod
    def artifact_path(self) -> Path:
        r"""Gets the path to the artifacts.

        Returns:
            ``pathlib.Path``: The path where you can write some
                artifacts related to the experiment.

        Raises:
            ``NotActivatedExpTrackerError`` if the experiment tracker
                is not activated.
        """

    @property
    @abstractmethod
    def checkpoint_path(self) -> Path:
        r"""Gets the path to the checkpoints.

        Returns:
            ``pathlib.Path``: The path where you can write the
                checkpoints.

        Raises:
            ``NotActivatedExpTrackerError`` if the experiment tracker
                is not activated.
        """

    @property
    @abstractmethod
    def experiment_id(self) -> str:
        r"""Gets the experiment ID.

        Returns:
            str: The experiment ID.

        Raises:
            ``NotActivatedExpTrackerError`` if the experiment tracker
                is not activated.
        """

    @abstractmethod
    def start(self) -> None:
        r"""Starts a new experiment tracking."""

    @abstractmethod
    def flush(self, upload_checkpoints: bool = True) -> None:
        r"""Flushes all the current artifacts/metrics to the remote server.

        This function is expected to be called not at a high frequency.
        For example, you can call this function at the end of each
        epoch.

        Args:
            upload_checkpoints (bool): Indicates if the checkpoints
                are uploaded or not when this method is called.
                Default: ``True``

        Raises:
            ``NotActivatedExpTrackerError`` if the experiment tracker
                is not activated.
        """

    @abstractmethod
    def end(self) -> None:
        r"""Ends the tracking of the current experiment.

        Raises:
            ``NotActivatedExpTrackerError`` if the experiment tracker
                is not activated.
        """

    @abstractmethod
    def is_activated(self) -> bool:
        r"""Indicates if the tracker is activated or not.

        Returns:
            bool: ``True`` if the tracker is activated, otherwise
                ``False``

        Raises:
            ``NotActivatedExpTrackerError`` if the experiment tracker
                is not activated.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.exp_trackers import BaseExpTracker
            >>> exp_tracker: BaseExpTracker = ...  # Initialize and activate the experiment tracker
            >>> exp_tracker.is_activated()
            True
        """

    @abstractmethod
    def is_resumed(self) -> bool:
        r"""Indicates if the experiment was resumed or not.

        Returns:
            bool: ``True`` if the experiment was resumed, otherwise
                ``False``

        Raises:
            ``NotActivatedExpTrackerError`` if the experiment tracker
                is not activated.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.exp_trackers import BaseExpTracker
            >>> exp_tracker: BaseExpTracker = ...  # Initialize and activate the experiment tracker
            >>> exp_tracker.is_resumed()
        """

    @abstractmethod
    def add_tag(self, name: str, value: Any) -> None:
        r"""Adds a tag to the experiment.

        Args:
            name (str): Specifies the name of the tag.
            value: Specifies the value of the tag. The value should be
                convertibled to a string  with the ``str`` function.

        Raises:
            ``NotActivatedExpTrackerError`` if the experiment tracker
                is not activated.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.exp_trackers import BaseExpTracker
            >>> exp_tracker: BaseExpTracker = ...  # Initialize and activate the experiment tracker
            >>> exp_tracker.add_tag('mode', 'training')
        """

    @abstractmethod
    def add_tags(self, tags: dict[str, Any]) -> None:
        r"""Adds tags to the experiment.

        Args:
            tags (dict): Specifies the tags to add to the experiment.
                The value should be convertibled to a string with the
                ``str`` function.

        Raises:
            ``NotActivatedExpTrackerError`` if the experiment tracker
                is not activated.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.exp_trackers import BaseExpTracker
            >>> exp_tracker: BaseExpTracker = ...  # Initialize and activate the experiment tracker
            >>> exp_tracker.add_tags({'mode': 'training', 'machine': 'mac'})
        """

    @abstractmethod
    def create_artifact(self, artifact: BaseArtifact) -> None:
        r"""Creates an artifact.

        Args:
            artifact (``BaseArtifact``): Specifies the artifact to
                create.

        Raises:
            ``NotActivatedExpTrackerError`` if the experiment tracker
                is not activated.

        Example usage:

        .. code-block:: python

            >>> exp_tracker: BaseExpTracker = ...  # Initialize and activate the experiment tracker
            >>> from gravitorch.utils.artifacts import JSONArtifact
            >>> exp_tracker.create_artifact(JSONArtifact(tag='metric', data={'f1_score': 42}))
        """

    @abstractmethod
    def log_best_metric(self, key: str, value: Any) -> None:
        r"""Logs the best value of a metric.

        Args:
            key (str): Specifies the key used to identify the metric.
                Please do not use the names `'epoch'`, `'iteration'`
                and `'step'` because they are reserved for specific
                uses.
            value (int or float): Specifies the best value to log.

        Raises:
            ``NotActivatedExpTrackerError`` if the experiment tracker
                is not activated.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.exp_trackers import BaseExpTracker
            >>> exp_tracker: BaseExpTracker = ...  # Initialize and activate the experiment tracker
            >>> exp_tracker.log_best_metric('my_metric', 1.2)
        """

    @abstractmethod
    def log_best_metrics(self, metrics: dict[str, Any]) -> None:
        r"""Logs a dictionary of the best metrics.

        Args:
            metrics (dict): Specifies the dictionary of the best
                metrics to log. Please do not use the keys
                ``'epoch'``,  ``'iteration'`` and ``'step'``
                because they are reserved for specific uses.

        Raises:
            ``NotActivatedExpTrackerError`` if the experiment tracker
                is not activated.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.exp_trackers import BaseExpTracker
            >>> exp_tracker: BaseExpTracker = ...  # Initialize and activate the experiment tracker
            >>> exp_tracker.log_best_metrics({"my_metric_1": 12, "my_metric_2": 3.5})
        """

    @abstractmethod
    def log_figure(self, key: str, figure: Figure, step: Optional[Step] = None) -> None:
        r"""Logs a figure for the given key and step.

        Args:
            key (str): Specifies the key used to identify the figure.
            figure (``matplotlib.pyplot.Figure``): Specifies the
                figure to log.
            step (``Step``, optional): Specifies the step value to
                record. If ``None``, it will use the default step.
                Default: ``None``

        Raises:
            ``NotActivatedExpTrackerError`` if the experiment tracker
                is not activated.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.exp_trackers import BaseExpTracker
            >>> exp_tracker: BaseExpTracker = ...  # Initialize and activate the experiment tracker
            >>> import matplotlib.pyplot as plt
            >>> fig, axes = plt.subplots()
            >>> exp_tracker.log_figure('my_figure', fig)  # without step
            >>> from gravitorch.utils.exp_trackers import EpochStep
            >>> exp_tracker.log_figure('my_figure', fig, EpochStep(2))  # with step
        """

    @abstractmethod
    def log_figures(self, figures: dict[str, Figure], step: Optional[Step] = None) -> None:
        r"""Logs a dictionary of figures for a given step.

        Args:
            figures (dict): Specifies the dictionary of figures to log.
            step (``Step``, optional): Specifies the step value to
                record. If ``None``, it will use the default step.
                Default: ``None``

        Raises:
            ``NotActivatedExpTrackerError`` if the experiment tracker
                is not activated.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.exp_trackers import BaseExpTracker
            >>> exp_tracker: BaseExpTracker = ...  # Initialize and activate the experiment tracker
            >>> import matplotlib.pyplot as plt
            >>> fig, axes = plt.subplots()
            >>> exp_tracker.log_figures({'my_figure_1': fig, 'my_figure_2': fig})  # without step
            >>> from gravitorch.utils.exp_trackers import EpochStep
            >>> exp_tracker.log_figures({'my_figure_1': fig, 'my_figure_2': fig}, EpochStep(2))  # with step
        """

    @abstractmethod
    def log_hyper_parameter(self, key: str, value: Any) -> None:
        r"""Logs a single hyper-parameter.

        Args:
            key (str): Specifies the name of the hyper-parameter.
            value (str): Specifies the value of the hyper-parameter.

        Raises:
            ``NotActivatedExpTrackerError`` if the experiment tracker
                is not activated.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.exp_trackers import BaseExpTracker
            >>> exp_tracker: BaseExpTracker = ...  # Initialize and activate the experiment tracker
            >>> exp_tracker.log_hyper_parameter('my_hparam', 32)
        """

    @abstractmethod
    def log_hyper_parameters(self, params: dict[str, Any]) -> None:
        r"""Logs a dictionary of multiple hyper-parameters.

        Args:
            params (dict): Specifies the hyper-parameters to log.

        Raises:
            ``NotActivatedExpTrackerError`` if the experiment tracker
                is not activated.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.exp_trackers import BaseExpTracker
            >>> exp_tracker: BaseExpTracker = ...  # Initialize and activate the experiment tracker
            >>> exp_tracker.log_hyper_parameters({'input_size': 32, 'output_size': 16})
        """

    @abstractmethod
    def log_image(self, key: str, image: Image, step: Optional[Step] = None) -> None:
        r"""Logs an image for the given key and step.

        Args:
            key (str): Specifies the key used to identify the image.
            image (``PIL.Image.Image``): Specifies the image to log.
            step (``Step``, optional): Specifies the step value to
                record. Default: ``None``.

        Raises:
            ``NotActivatedExpTrackerError`` if the experiment tracker
                is not activated.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.exp_trackers import BaseExpTracker
            >>> exp_tracker: BaseExpTracker = ...  # Initialize and activate the experiment tracker
            >>> from PIL import Image
            >>> import numpy as np
            >>> img = Image.fromarray(np.zeros((256, 256, 3), dtype=np.uint8), 'RGB')
            >>> exp_tracker.log_image('my_image', img)  # without step
            >>> from gravitorch.utils.exp_trackers import EpochStep
            >>> exp_tracker.log_image('my_image', img, EpochStep(2))  # with step
        """

    @abstractmethod
    def log_images(self, images: dict[str, Image], step: Optional[Step] = None) -> None:
        r"""Logs a dictionary of images for a given step.

        Args:
            images (dict): Specifies the dictionary of images to log.
            step (``Step``, optional): Specifies the step value to
                record. If ``None``, it will use the default step.
                Default: ``None``.

        Raises:
            ``NotActivatedExpTrackerError`` if the experiment tracker
                is not activated.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.exp_trackers import BaseExpTracker
            >>> exp_tracker: BaseExpTracker = ...  # Initialize and activate the experiment tracker
            >>> from PIL import Image
            >>> import numpy as np
            >>> img = Image.fromarray(np.zeros((256, 256, 3), dtype=np.uint8), 'RGB')
            >>> exp_tracker.log_images({'my_image_1': img, 'my_image_2': img})  # without step
            >>> from gravitorch.utils.exp_trackers.steps import EpochStep
            >>> exp_tracker.log_images({'my_image_1': img, 'my_image_2': img}, EpochStep(2))  # with step
        """

    @abstractmethod
    def log_metric(self, key: str, value: Union[int, float], step: Optional[Step] = None) -> None:
        r"""Logs a single metric.

        Args:
            key (str): Specifies the key used to identify the metric.
                Please do not use the names ``'epoch'``,
                ``'iteration'`` and ``'step'`` because they are
                reserved for specific uses.
            value (int or float): Specifies the value to log.
            step (``Step``, optional): Specifies the step value to
                record. If ``None``, it will use the default step.
                Default: ``None``.

        Raises:
            ``NotActivatedExpTrackerError`` if the experiment tracker
                is not activated.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.exp_trackers import BaseExpTracker
            >>> exp_tracker: BaseExpTracker = ...  # Initialize and activate the experiment tracker
            >>> exp_tracker.log_metric('my_metric', 1.2)  # without step
            >>> from gravitorch.utils.exp_trackers import EpochStep
            >>> exp_tracker.log_metric('my_metric', 1.2, EpochStep(2))  # with step
        """

    @abstractmethod
    def log_metrics(
        self, metrics: dict[str, Union[int, float]], step: Optional[Step] = None
    ) -> None:
        r"""Logs a dictionary of multiple metrics.

        Args:
            metrics (dict): Specifies the dictionary of metrics to log.
                Please do not use the keys ``'epoch'``, ``'iteration'``
                and ``'step'`` because they are reserved for specific
                uses.
            step (``Step``, optional): Specifies the step value to
                record. If ``None``, it will use the default step.
                Default: ``None``.

        Raises:
            ``NotActivatedExpTrackerError`` if the experiment tracker
                is not activated.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.exp_trackers import BaseExpTracker
            >>> exp_tracker: BaseExpTracker = ...  # Initialize and activate the experiment tracker
            >>> exp_tracker.log_metrics({'my_metric_1': 1.2, 'my_metric_2': 42})  # without step
            >>> from gravitorch.utils.exp_trackers import EpochStep
            >>> exp_tracker.log_metrics({'my_metric_1': 1.2, 'my_metric_2': 42}, EpochStep(2))  # with step
        """

    @abstractmethod
    def upload_checkpoints(self) -> None:
        r"""Uploads all the checkpoints that are in the checkpoint path to the
        remote server.

        This method does nothing if no remote server is used.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.exp_trackers import BaseExpTracker
            >>> exp_tracker: BaseExpTracker = ...  # Initialize and activate the experiment tracker
            >>> exp_tracker.upload_checkpoints()
        """


class NotActivatedExpTrackerError(Exception):
    r"""This exception is raised when you try to do an action with an
    experiment tracker which is not activated."""


class BaseBasicExpTracker(BaseExpTracker):
    r"""Defines a base class to implement some experiment trackers while
    avoiding duplication for some methods.

    This base class implements some basic functionalities to avoid
    duplicate code e.g. code to raise an error if the experiment
    tracker is not implemented.

    Note: it is not necessary to use this base class to implement an
    experiment tracker.
    """

    def start(self) -> None:
        if self.is_activated():
            raise RuntimeError("The experiment tracker is already activated")
        self._start()

    @abstractmethod
    def _start(self) -> None:
        r"""Starts a new experiment tracking."""

    def flush(self, upload_checkpoints: bool = True) -> None:
        if not self.is_activated():
            raise NotActivatedExpTrackerError(
                "It is not possible to create flush data because the experiment tracker is not activated"
            )
        self._flush(upload_checkpoints)

    @abstractmethod
    def _flush(self, upload_checkpoints: bool = True) -> None:
        r"""Flushes all the current artifacts/metrics to the remote server.

        This function is expected to be called not at a high frequency.
        For example, you can call this function at the end of each
        epoch.

        Args:
            upload_checkpoints (bool, optional): Indicates if the
                checkpoints are uploaded or not when this method is
                called. Default: ``True``
        """

    def end(self) -> None:
        if not self.is_activated():
            raise NotActivatedExpTrackerError(
                "It is not possible to end the tracking because the experiment tracker "
                "is not activated"
            )
        self._end()

    def _end(self) -> None:
        r"""Ends the tracking of the current experiment."""

    def is_resumed(self) -> bool:
        if not self.is_activated():
            raise NotActivatedExpTrackerError(
                "It is not possible to indicate if the experiment was resumed "
                "or not because the experiment tracker is not activated"
            )
        return self._is_resumed()

    @abstractmethod
    def _is_resumed(self) -> bool:
        r"""Indicates if the experiment was resumed or not.

        Returns:
            bool: ``True`` if the experiment was resumed, otherwise
                ``False``.
        """

    def add_tag(self, name: str, value: Any) -> None:
        if not self.is_activated():
            raise NotActivatedExpTrackerError(
                f"It is not possible to create add the tag {name} because the experiment "
                "tracker is not activated"
            )
        self._add_tag(name, value)

    @abstractmethod
    def _add_tag(self, name: str, value: Any) -> None:
        r"""Adds a tag to the experiment.

        Args:
            name (str): Specifies the name of the tag.
            value: Specifies the value of the tag. The value should be
                convertible to a string with the ``str`` function.
        """

    def add_tags(self, tags: dict[str, Any]) -> None:
        if not self.is_activated():
            raise NotActivatedExpTrackerError(
                "It is not possible to create add tags because the experiment tracker is "
                "not activated"
            )
        self._add_tags(tags)

    @abstractmethod
    def _add_tags(self, tags: dict[str, Any]) -> None:
        r"""Adds tags to the experiment.

        Args:
            tags (dict): Specifies the tags to add to the experiment.
                The value should be convertible to a string with the
                ``str`` function.
        """

    def create_artifact(self, artifact: BaseArtifact) -> None:
        if not self.is_activated():
            raise NotActivatedExpTrackerError(
                f"It is not possible to create the artifact {artifact} because "
                "the experiment tracker is not activated."
            )
        self._create_artifact(artifact)

    @abstractmethod
    def _create_artifact(self, artifact: BaseArtifact) -> None:
        r"""Creates an artifact.

        Args:
            artifact (``BaseArtifact``): Specifies the artifact to
                create.
        """

    def log_best_metric(self, key: str, value: Any) -> None:
        if not self.is_activated():
            raise NotActivatedExpTrackerError(
                f"It is not possible to log the best metric for {key} because "
                "the experiment tracker is not activated."
            )
        self._log_best_metric(key, value)

    @abstractmethod
    def _log_best_metric(self, key: str, value: Any) -> None:
        r"""Logs the best value of a metric.

        Args:
            key (str): Specifies the key used to identify the metric.
                Please do not use the names ``'epoch'``,
                ``'iteration'`` and ``'step'`` because they are
                reserved for specific uses.
            value (int or float): Specifies the best value to log.
        """

    def log_best_metrics(self, metrics: dict[str, Any]) -> None:
        if not self.is_activated():
            raise NotActivatedExpTrackerError(
                "It is not possible to log the best metrics because the experiment tracker "
                "is not activated"
            )
        self._log_best_metrics(metrics)

    @abstractmethod
    def _log_best_metrics(self, metrics: dict[str, Any]) -> None:
        r"""Logs a dictionary of the best metrics.

        Args:
            metrics (dict): Specifies the dictionary of the best
                metrics to log. Please do not use the keys ``'epoch'``,
                ``'iteration'``, and ``'step'`` because they are
                reserved for specific uses.
        """

    def log_figure(self, key: str, figure: Figure, step: Optional[Step] = None) -> None:
        if not self.is_activated():
            raise NotActivatedExpTrackerError(
                f"It is not possible to log the figure {key} because the experiment tracker "
                "is not activated"
            )
        self._log_figure(key, figure, step)

    @abstractmethod
    def _log_figure(self, key: str, figure: Figure, step: Optional[Step] = None) -> None:
        r"""Logs a figure for the given key and step.

        Args:
            key (str): Specifies the key used to identify the figure.
            figure (``matplotlib.pyplot.Figure``): Specifies the
                figure to log.
            step (``Step``, optional): Specifies the step value to
                record. If ``None``, it will use the default step.
                Default: ``None``
        """

    def log_figures(self, figures: dict[str, Figure], step: Optional[Step] = None) -> None:
        if not self.is_activated():
            raise NotActivatedExpTrackerError(
                "It is not possible to log the figures because the experiment tracker "
                "is not activated"
            )
        self._log_figures(figures, step)

    @abstractmethod
    def _log_figures(self, figures: dict[str, Figure], step: Optional[Step] = None) -> None:
        r"""Logs a dictionary of figures for a given step.

        Args:
            figures (dict): Specifies the dictionary of figures to log.
            step (``Step``, optional): Specifies the step value to
                record. If ``None``, it will use the default step.
                Default: ``None``
        """

    def log_hyper_parameter(self, key: str, value: Any) -> None:
        if not self.is_activated():
            raise NotActivatedExpTrackerError(
                f"It is not possible to log the hyper-parameter {key} because the "
                "experiment tracker is not activated"
            )
        self._log_hyper_parameter(key, value)

    @abstractmethod
    def _log_hyper_parameter(self, key: str, value: Any) -> None:
        r"""Logs a single hyper-parameter.

        Args:
            key (str): Specifies the name of the hyper-parameter.
            value (str): Specifies the value of the hyper-parameter.
        """

    def log_hyper_parameters(self, params: dict[str, Any]) -> None:
        if not self.is_activated():
            raise NotActivatedExpTrackerError(
                "It is not possible to log the hyper-parameters because the experiment tracker "
                "is not activated"
            )
        self._log_hyper_parameters(params)

    @abstractmethod
    def _log_hyper_parameters(self, params: dict[str, Any]) -> None:
        r"""Logs a dictionary of multiple hyper-parameters.

        Args:
            params (dict): Specifies the hyper-parameters to log.
        """

    def log_image(
        self,
        key: str,
        image: Image,  # noqa: F821
        step: Optional[Step] = None,
    ) -> None:
        if not self.is_activated():
            raise NotActivatedExpTrackerError(
                f"It is not possible to log the image {key} because the experiment tracker "
                "is not activated"
            )
        self._log_image(key, image, step)

    @abstractmethod
    def _log_image(
        self,
        key: str,
        image: Image,  # noqa: F821
        step: Optional[Step] = None,
    ) -> None:
        r"""Logs an image for the given key and step.

        Args:
            key (str): Specifies the key used to identify the image.
            image (``PIL.Image.Image``): Specifies the image to log.
            step (``Step``, optional): Specifies the step value to
                record. If ``None``, it will use the default step.
                Default: ``None``.
        """

    def log_images(self, images: dict[str, Image], step: Optional[Step] = None) -> None:
        if not self.is_activated():
            raise NotActivatedExpTrackerError(
                "It is not possible to log the images because the experiment tracker "
                "is not activated"
            )
        self._log_images(images, step)

    @abstractmethod
    def _log_images(self, images: dict[str, Image], step: Optional[Step] = None) -> None:
        r"""Logs a dictionary of images for a given step.

        Args:
            images (dict): Specifies the dictionary of images to log.
            step (``Step``, optional): Specifies the step value to
                record. If ``None``, it will use the default step.
                Default: ``None``.
        """

    def log_metric(self, key: str, value: Union[int, float], step: Optional[Step] = None) -> None:
        if not self.is_activated():
            raise NotActivatedExpTrackerError(
                f"It is not possible to log metric {key} because the experiment tracker "
                "is not activated"
            )
        self._log_metric(key, value, step)

    @abstractmethod
    def _log_metric(self, key: str, value: Union[int, float], step: Optional[Step] = None) -> None:
        r"""Logs a single metric.

        Args:
            key (str): Specifies the key used to identify the metric.
                Please do not use the names ``'epoch'``,
                ``'iteration'`` and ``'step'`` because they are
                reserved for specific uses.
            value (int or float): Specifies the value to log.
            step (``Step``, optional): Specifies the step value to
                record. If ``None``, it will use the default step.
                Default: ``None``.
        """

    def log_metrics(
        self, metrics: dict[str, Union[int, float]], step: Optional[Step] = None
    ) -> None:
        if not self.is_activated():
            raise NotActivatedExpTrackerError(
                "It is not possible to log metrics because the experiment tracker is not activated"
            )
        self._log_metrics(metrics, step)

    @abstractmethod
    def _log_metrics(
        self, metrics: dict[str, Union[int, float]], step: Optional[Step] = None
    ) -> None:
        r"""Logs a dictionary of multiple metrics.

        Args:
            metrics (dict): Specifies the dictionary of metrics to log.
                Please do not use the keys ``'epoch'``,
                ``'iteration'`` and ``'step'`` because they are
                reserved for specific uses.
            step (``Step``, optional): Specifies the step value to
                record. If ``None``, it will use the default step.
                Default: ``None``.
        """

    def upload_checkpoints(self) -> None:
        if not self.is_activated():
            raise NotActivatedExpTrackerError(
                "It is not possible to upload checkpoints because the experiment tracker "
                "is not activated"
            )
        self._upload_checkpoints()

    @abstractmethod
    def _upload_checkpoints(self) -> None:
        r"""Uploads all the checkpoints to the remote server.

        This method does nothing if no remote server is used.
        """



__all__ = ["ModelParameterAnalyzer"]

from typing import Union

from gravitorch.engines.base import BaseEngine
from gravitorch.engines.events import EngineEvents
from gravitorch.handlers.base import BaseHandler
from gravitorch.handlers.utils import to_events
from gravitorch.nn.utils.parameter_analysis import show_parameter_stats
from gravitorch.utils.events import VanillaEventHandler


class ModelParameterAnalyzer(BaseHandler):
    r"""Implements a handler to analyze the model parameter values.

    Args:
        events (str or tuple or list): Specifies the event(s) when to
            analyze the model parameter values. For example, it is
            usually a good idea to look at the model parameters at the
            beginning and end of the training.
            Default: ``('started', 'train_completed')``
        tablefmt (str, optional): Specifies the table format. You can
            find the valid formats at
            https://pypi.org/project/tabulate/. Default: ``'rst'``
    """

    def __init__(
        self,
        events: Union[str, tuple[str, ...], list[str]] = (
            EngineEvents.STARTED,
            EngineEvents.TRAIN_COMPLETED,
        ),
        tablefmt: str = "rst",
    ):
        self._events = to_events(events)
        self._tablefmt = tablefmt

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(events={self._events}, tablefmt={self._tablefmt})"

    def attach(self, engine: BaseEngine) -> None:
        for event in self._events:
            engine.add_event_handler(
                event,
                VanillaEventHandler(self.analyze, handler_kwargs={"engine": engine}),
            )

    def analyze(self, engine: BaseEngine) -> None:
        r"""Analyzes the model parameter values.

        Args:
            engine (``BaseEngine``): Specifies the engine with the
                model to analyze.
        """
        show_parameter_stats(module=engine.model, tablefmt=self._tablefmt)



__all__ = ["exponential_nll_loss"]

from torch import Tensor

from gravitorch.nn.functional.loss_helpers import basic_loss_reduction


def exponential_nll_loss(
    log_rate: Tensor,
    target: Tensor,
    log_input: bool = True,
    eps: float = 1e-8,
    max_log_value: float = 20.0,
    reduction: str = "mean",
) -> Tensor:
    r"""Computes the negative log-likelihood (NLL) with Exponential distribution
    of target.

    Args:
        log_rate (``torch.Tensor`` of type float): Specifies the
            predicted rates (``lambda``).
        target (``torch.Tensor`` of type float and same shape as
            ``log_rate``): Specifies the target values.
        log_input (bool, optional): If ``True``, the expected input
            is ``log(rate)``, otherwise it is ``rate``.
            Default: ``True``
        eps (float, optional): Small value to avoid evaluation of
            :math:`\log(0)` when :attr:`log_input = False`.
            Default: ``1e-8``
        max_log_value (float, optional): Specifies the maximum value
            used to clip ``log_rate`` before to compute the
            exponential when :attr:`log_input = True`.
            Default: ``20.0``
        reduction (string, optional): Specifies the reduction to
            apply to the output: ``'none'`` | ``'mean'`` | ``'sum'``.
            ``'none'``: no reduction will be applied, ``'mean'``: the
            sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be
            summed. Default: ``'mean'``

    Returns:
        ``torch.Tensor`` of type float: The negative log-likelihood
            with Exponential distribution of target. The shape of the
            tensor depends on the reduction strategy.
    """
    if log_input:
        # Input is log(rate)
        nll = log_rate.clamp(max=max_log_value).exp().mul(target) - log_rate
    else:
        # Input is rate
        nll = log_rate.mul(target) - log_rate.add(eps).log()
    return basic_loss_reduction(nll, reduction)



r"""This module implements some meters."""

from __future__ import annotations

__all__ = ["ScalarMetricTracker"]

import logging
import math
from collections import defaultdict
from typing import TYPE_CHECKING

import torch

from gravitorch.utils.exp_trackers.steps import EpochStep
from gravitorch.utils.format import str_scalar
from gravitorch.utils.meters.scalar import ScalarMeter

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

logger = logging.getLogger(__name__)


class ScalarMetricTracker:
    r"""Implements a tracker for multiple scalar metrics.

    This tracker use the ``ScalarMeter`` to track some values of each
    metric.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.metric_tracker import ScalarMetricTracker
        >>> tracker = ScalarMetricTracker()
        >>> tracker.update({'metric1': 1, 'metric2': 2})
        >>> tracker.update({'metric1': 11, 'metric2': 12})
        >>> tracker.update({'metric1': 21, 'metric2': 22})
        >>> tracker.log_average_value()
        INFO:gravitorch.utils.metric_tracker:metric1: 11.0000
        INFO:gravitorch.utils.metric_tracker:metric2: 12.0000
    """

    def __init__(self):
        self._metrics = defaultdict(ScalarMeter)

    def update(self, data: dict) -> None:
        r"""Tracks the values that can be convert in int or float.

        The values that cannot be convert in int or float are ignored.
        Similarly, the NaN are ignored.

        Args:
            data (dict): Specifies a dict with the metric values to
                track.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.metric_tracker import ScalarMetricTracker
            >>> tracker = ScalarMetricTracker()
            >>> tracker.update({'metric1': 1, 'metric2': 2})
        """
        for key, value in data.items():
            if torch.is_tensor(value) and value.numel() == 1:
                value = value.item()  # get number from a torch scalar
            if isinstance(value, (int, float)) and not math.isnan(value):
                self._metrics[key].update(value)

    def log_average_value(self, engine: BaseEngine | None = None, prefix: str = "") -> None:
        r"""Logs the average value of the metrics to the engine.

        Args:
            engine (``BaseEngine`` or None, optional): Specifies the
                engine. Default: ``None``
            prefix (str, optional): Specifies the prefix used to log
                the metric name. Default: ``''``

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.metric_tracker import ScalarMetricTracker
            >>> tracker = ScalarMetricTracker()
            >>> tracker.update({'metric1': 1, 'metric2': 2})
            >>> tracker.log_average_value()
            INFO:gravitorch.utils.metric_tracker:metric1: 1.000000
            INFO:gravitorch.utils.metric_tracker:metric2: 2.000000
            >>> tracker.log_average_value(prefix='train/')
            INFO:gravitorch.utils.metric_tracker:train/metric1: 1.000000
            INFO:gravitorch.utils.metric_tracker:train/metric2: 2.000000
        """
        metrics = {f"{prefix}{key}": value.average() for key, value in self._metrics.items()}
        if engine:
            engine.log_metrics(metrics, step=EpochStep(engine.epoch))
        for name, value in metrics.items():
            logger.info(f"{name}: {str_scalar(value)}")



r"""This file implements a module manager."""

__all__ = ["ModuleManager"]

import logging
from typing import Any, Union

from gravitorch.utils.format import str_add_indent, to_torch_mapping_str

logger = logging.getLogger(__name__)


class ModuleManager:
    r"""Implements a module manager to easily manage a group of modules.

    This module manager assumes that the modules have a ``state_dict``
    and ``load_state_dict`` methods.
    """

    def __init__(self):
        self._modules = {}

    def __len__(self) -> int:
        return len(self._modules)

    def __repr__(self) -> str:
        if self._modules:
            return (
                f"{self.__class__.__qualname__}(\n"
                f"  {str_add_indent(to_torch_mapping_str(self._modules))}\n)"
            )
        return f"{self.__class__.__qualname__}()"

    def add_module(self, name: str, module: Any) -> None:
        r"""Adds a module to the module state manager.

        Note that the name should be unique. If the name exists, the
        old module will be overwritten by the new module.

        Args:
            name (str): Specifies the name of the module to add.
            module: Specifies the module to add.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.module_manager import ModuleManager
            >>> manager = ModuleManager()
            >>> from torch import nn
            >>> manager.add_module('my_module', nn.Linear(4, 6))
        """
        if name in self._modules:
            logger.info(f"Overriding the '{name}' module")
        self._modules[name] = module

    def get_module(self, name: str) -> Any:
        r"""Gets a module.

        Args:
            name (str): Specifies the module to get.

        Returns:
            The module

        Raises:
            ValueError if the module does not exist.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.module_manager import ModuleManager
            >>> manager = ModuleManager()
            >>> from torch import nn
            >>> manager.add_module('my_module', nn.Linear(4, 6))
            >>> manager.get_module('my_module')
            nn.Linear(4, 6)
        """
        if not self.has_module(name):
            raise ValueError(f"The module {name} does not exist")
        return self._modules[name]

    def has_module(self, name: str) -> bool:
        r"""Indicates if there is module for the given name.

        Args:
            name (str): Specifies the name to check.

        Returns:
            bool: ``True`` if the module exists, otherwise ``False``

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.module_manager import ModuleManager
            >>> manager = ModuleManager()
            >>> from torch import nn
            >>> manager.add_module('my_module', nn.Linear(4, 6))
            >>> manager.has_module('my_module')
            True
            >>> manager.has_module('missing')
            False
        """
        return name in self._modules

    def remove_module(self, name: str) -> None:
        r"""Removes a module.

        Args:
            name (str): Specifies the name of the module to remove.

        Raises:
            ValueError if the module does not exist.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.module_manager import ModuleManager
            >>> manager = ModuleManager()
            >>> from torch import nn
            >>> manager.add_module('my_module', nn.Linear(4, 6))
            >>> manager.remove_module('my_module')
            >>> manager.has_module('my_module')
            False
        """
        if name not in self._modules:
            raise ValueError(f"The module {name} does not exist so it is not possible to remove it")
        self._modules.pop(name)

    def load_state_dict(self, state_dict: dict, keys: Union[list, tuple, None] = None) -> None:
        r"""Loads the state dict of each module.

        Note this method ignore the missing modules or keys. For
        example if you want to load the optimizer module but there is
        no 'optimizer' key in the state dict, this method will ignore
        the optimizer module.

        Args:
            state_dict (dict): Specifies the state dict to load.
            keys (list or tuple or ``None``): Specifies the keys to
                load. If ``None``, it loads all the keys associated
                to the registered modules.
        """
        keys = keys or tuple(self._modules.keys())
        for key in keys:
            if key not in state_dict:
                logger.info(f"Ignore key {key} because it is not in the state dict")
                continue
            if key not in self._modules:
                logger.info(f"Ignore key {key} because there is no module associated to it")
                continue
            if not hasattr(self._modules[key], "load_state_dict"):
                logger.info(
                    f"Ignore key {key} because the module does not have 'load_state_dict' method"
                )
                continue
            self._modules[key].load_state_dict(state_dict[key])

    def state_dict(self) -> dict:
        r"""Creates a state dict with all the modules.

        The state of each module is store with the associated key of
        the module.

        Returns:
            dict: The state dict of all the modules.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.module_manager import ModuleManager
            >>> manager = ModuleManager()
            >>> from torch import nn
            >>> manager.add_module('my_module', nn.Linear(4, 6))
            >>> manager.state_dict()
            {'my_module': OrderedDict([('weight', tensor([[-0.3810, -0.3761,  0.3272,  0.2582],
                    [ 0.4013,  0.0173,  0.3931, -0.1642],
                    [-0.4772,  0.1962, -0.0101,  0.0654],
                    [-0.1067,  0.3788, -0.0949, -0.3952],
                    [ 0.0111, -0.2536, -0.3626, -0.0810],
                    [-0.1757, -0.4256, -0.1076, -0.2050]])),
                ('bias', tensor([ 0.2897, -0.4282, -0.2547, -0.2704, -0.1545,  0.2825]))])}
            >>> manager.add_module('int', 123)
            >>> manager.state_dict()
            {'my_module': OrderedDict([('weight', tensor([[-0.3810, -0.3761,  0.3272,  0.2582],
                    [ 0.4013,  0.0173,  0.3931, -0.1642],
                    [-0.4772,  0.1962, -0.0101,  0.0654],
                    [-0.1067,  0.3788, -0.0949, -0.3952],
                    [ 0.0111, -0.2536, -0.3626, -0.0810],
                    [-0.1757, -0.4256, -0.1076, -0.2050]])),
                ('bias', tensor([ 0.2897, -0.4282, -0.2547, -0.2704, -0.1545,  0.2825]))])}
        """
        state = {}
        for name, module in self._modules.items():
            if hasattr(module, "state_dict"):
                state[name] = module.state_dict()
            else:
                logger.info(f"Skip '{name}' module because it does not have 'state_dict' method")
        return state



__all__ = ["SequentialMetric"]

from collections.abc import Sequence
from typing import Optional, Union

from torch import nn

from gravitorch.engines.base import BaseEngine
from gravitorch.models.metrics.base import BaseMetric, setup_metric


class SequentialMetric(BaseMetric):
    r"""Implements a "metric" to evaluate a sequence of metrics.

    This metric assumes that all the metrics are consistent. All the
    metrics should be compatible with the same inputs. The outputs
    are combined so the keys returned by each metric should be
    different.

    Args:
        metrics (sequence): Specifies the list of metrics to use or
            their configurations.
    """

    def __init__(self, metrics: Sequence[Union[BaseMetric, dict]]):
        super().__init__()
        self.metrics = nn.ModuleList([setup_metric(metric) for metric in metrics])

    def attach(self, engine: BaseEngine) -> None:
        r"""Attaches the metrics to the provided engine.

        Note that each metric is attached individually so the metric
        handlers can be triggered at different events. This method
        calls the ``attach`` method of each metric. This method can
        be used to:

            - add event handler to the engine
            - set up history trackers

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        for metric in self.metrics:
            metric.attach(engine)

    def forward(self, *args, **kwargs) -> dict:
        r"""Updates the metrics given a mini-batch of examples.

        This method calls the ``forward`` method of each metric and
        combines the outputs in a single dictionary. All the metrics
        should be compatible with the inputs.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            dict: The outputs of the metrics.
        """
        outputs = {}
        for metric in self.metrics:
            out = metric(*args, **kwargs)
            if out:
                outputs.update(out)
        return outputs

    def reset(self) -> None:
        r"""Resets all the metrics."""
        for metric in self.metrics:
            metric.reset()

    def value(self, engine: Optional[BaseEngine] = None) -> dict:
        r"""Evaluates all the metrics and log the results given all the examples
        previously seen.

        This method calls the ``value`` method of each metric and
        combines the outputs in a single dictionary.

        Args:
            engine (``BaseEngine``, optional): Specifies the engine.
                This argument is required to log the results in the
                engine. Default: ``None``.

        Returns:
             dict: The results of the metric
        """
        outputs = {}
        for metric in self.metrics:
            outputs.update(metric.value(engine))
        return outputs



r"""This module contains some utility functions around ``torch.device``."""

__all__ = ["get_available_devices", "move_to_device"]

from typing import TypeVar

import torch
from torch.nn.utils.rnn import PackedSequence

T = TypeVar("T")


def get_available_devices() -> tuple[str, ...]:
    r"""Gets the available PyTorch devices on the machine.

    Returns:
        tuple: The available devices.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.torch_device import get_available_devices
        >>> get_available_devices()
        ('cpu', 'cuda:0')
    """
    if torch.cuda.is_available():
        return ("cpu", "cuda:0")
    return ("cpu",)


def move_to_device(data: T, device: torch.device) -> T:
    r"""Moves an object to a given device.

    If the object is a nested object (e.g. list, tuple, dictionary,
    set), this function sends the elements to the device. The current
    implementation supports the following types:

        - ``collections.OrderedDict``
        - ``dict``
        - ``list``
        - ``set``
        - ``torch.Tensor``
        - ``torch.nn.Module``
        - ``torch.nn.utils.rnn.PackedSequence``
        - ``tuple``

    Based on https://github.com/huggingface/accelerate

    Args:
        data: Specifies the data to move to the device. If it is a
            nested object, the data is moved recursively to the
            device.
        device (``torch.device``): Specifies the device to send the
            data to.

    Returns:
        The object on the given device. If it is not possible to move
            the object to the device, the input object is returned.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.utils.torch_device import move_to_device
        >>> move_to_device(
        ...     {'tensor1': torch.ones(2, 3), 'tensor2': torch.zeros(4)},
        ...     device=torch.device('cuda:0'),
        ... )
        {'tensor1': tensor([[1., 1., 1.], [1., 1., 1.]], device='cuda:0'),
         'tensor2': tensor([0., 0., 0., 0.], device='cuda:0')}
    """
    if isinstance(data, PackedSequence):
        return data.to(device)
    if isinstance(data, (list, tuple, set)):
        return type(data)(move_to_device(t, device) for t in data)
    if isinstance(data, dict):
        return type(data)({k: move_to_device(v, device) for k, v in data.items()})
    if not hasattr(data, "to"):
        return data
    return data.to(device)



__all__ = ["HypercubeVertexDataCreator"]

from typing import Optional

import torch

from gravitorch import constants as ct
from gravitorch.data.datacreators.base import BaseDataCreator
from gravitorch.engines.base import BaseEngine
from gravitorch.utils.seed import get_torch_generator


class HypercubeVertexDataCreator(BaseDataCreator[dict]):
    r"""Implements a data creator to create a toy classification dataset.

    The data are generated by using a hypercube. The targets are some
    vertices of the hypercube. Each input feature is a 1-hot
    representation of the target plus a Gaussian noise. These data can
    be used for a multi-class classification task.

    Args:
        num_examples (int, optional): Specifies the number of examples.
            Default: ``1000``
        num_classes (int, optional): Specifies the number of classes.
            Default: 50
        feature_size (int, optional): Specifies the feature size.
            The feature size has to be greater than the number of
            classes. Default: ``64``
        noise_std (float, optional): Specifies the standard deviation
            of the Gaussian noise. Default: ``0.2``
        random_seed (int, optional): Specifies the random seed used to
            initialize a ``torch.Generator`` object.
            Default: ``15782179921860610490``
    """

    def __init__(
        self,
        num_examples: int = 1000,
        num_classes: int = 50,
        feature_size: int = 64,
        noise_std: float = 0.2,
        random_seed: int = 15782179921860610490,
    ):
        if num_examples < 1:
            raise ValueError(f"The number of examples ({num_examples}) has to be greater than 0")
        self._num_examples = int(num_examples)

        if num_classes < 1:
            raise ValueError(f"The number of classes ({num_classes}) has to be greater than 0")
        self._num_classes = int(num_classes)

        if feature_size < num_classes:
            raise ValueError(
                f"The feature dimension ({feature_size:,}) has to be greater or equal to the "
                f"number of classes ({num_classes:,})"
            )
        self._feature_size = int(feature_size)

        if noise_std < 0:
            raise ValueError(
                f"The standard deviation of the Gaussian noise ({noise_std}) has to be "
                "greater or equal than 0"
            )
        self._noise_std = float(noise_std)

        self._torch_rng = get_torch_generator(random_seed)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}("
            f"num_examples={self._num_examples:,}, "
            f"num_classes={self._num_classes:,}, "
            f"feature_size={self._feature_size:,}, "
            f"noise_std={self._noise_std:,}, "
            f"random_seed={self.random_seed})"
        )

    @property
    def num_examples(self) -> int:
        r"""int: The number of examples when the data are created."""
        return self._num_examples

    @property
    def num_classes(self) -> int:
        r"""int: The number of classes when the data are created."""
        return self._num_classes

    @property
    def feature_size(self) -> int:
        r"""int: The feature size when the data are created."""
        return self._feature_size

    @property
    def noise_std(self) -> float:
        r"""float: The standard deviation of the Gaussian noise."""
        return self._noise_std

    @property
    def random_seed(self) -> int:
        r"""int: The random seed used to initialize a ``torch.Generator`` object."""
        return self._torch_rng.initial_seed()

    def create(self, engine: Optional[BaseEngine] = None) -> dict:
        r"""Creates data.

        Args:
            engine (``BaseEngine`` or ``None``): Specifies an engine.
                This input is not used in this data creator.
                Default: ``None``

        Returns:
            dict: A dictionary with two keys:

                - ``'input'``: a ``torch.Tensor`` of type float and
                    shape ``(num_examples, feature_size)``. This
                    tensor represents the input features.
                - ``'target'``: a ``torch.Tensor`` of type long and
                    shape ``(num_examples,)``. This tensor represents
                    the targets.
        """
        # Generate the target of each example.
        targets = torch.randint(
            0, self._num_classes, (self._num_examples,), generator=self._torch_rng
        )
        # Generate the features. Each class should be a vertex of the hyper-cube plus some noise.
        features = torch.randn(
            self._num_examples, self._feature_size, generator=self._torch_rng
        ).mul(self._noise_std)
        features.scatter_add_(
            1, targets.view(self._num_examples, 1), torch.ones(self._num_examples, 1)
        )
        return {ct.TARGET: targets, ct.INPUT: features}



__all__ = ["IterDataPipeCreatorDataSource", "DataCreatorIterDataPipeCreatorDataSource"]

import logging
from collections.abc import Iterable
from typing import Any, Optional, TypeVar, Union

from torch.utils.data import IterDataPipe

from gravitorch.creators.datapipe import (
    BaseIterDataPipeCreator,
    setup_iter_datapipe_creator,
)
from gravitorch.data.datacreators import BaseDataCreator, setup_data_creator
from gravitorch.datasources.base import BaseDataSource, LoaderNotFoundError
from gravitorch.engines.base import BaseEngine
from gravitorch.utils.asset_manager import AssetManager
from gravitorch.utils.format import str_add_indent, to_torch_mapping_str
from gravitorch.utils.summary import concise_summary

logger = logging.getLogger(__name__)

T = TypeVar("T")


class IterDataPipeCreatorDataSource(BaseDataSource):
    r"""Implements a data source that creates data loaders using
    ``IterDataPipe`` creators.

    Args:
        datapipe_creators (dict): Specifies the ``IterDataPipe``
            creators. Each key is associated to a loader ID. For
            example if you want to use a ``'train'`` data loader,
            you need to have a key associated to a
            ``BaseIterDataPipeCreator`` object or its configuration.
            Each ``BaseIterDataPipeCreator`` object contains the
            recipe to create an ``IterDataPipe`` object.

    Example:

    .. code-block:: python

        >>> from gravitorch.datasources import IterDataPipeCreatorDataSource
        # Create by using a ``BaseIterDataPipeCreator`` object.
        >>> from gravitorch.creators.datapipe import SequentialIterDataPipeCreator
        >>> from gravitorch.data.datapipes.iter import SourceWrapperIterDataPipe
        >>> data_source = IterDataPipeCreatorDataSource(
        ...     datapipe_creators={
        ...         "train": SequentialIterDataPipeCreator(
        ...             config=[
        ...                 {
        ...                     '_target_': "gravitorch.data.datapipes.iter.SourceWrapperIterDataPipe",
        ...                     "data": [1, 2, 3, 4],
        ...                 },
        ...             ]
        ...         ),
        ...         "val": SequentialIterDataPipeCreator(
        ...             config=[
        ...                 {
        ...                     '_target_': "gravitorch.data.datapipes.iter.SourceWrapperIterDataPipe",
        ...                     "data": ["a", "b", "c"],
        ...                 },
        ...             ]
        ...         ),
        ...     }
        ... )
        # Create by using the configs
        >>> from objectory import OBJECT_TARGET
        >>> data_source = IterDataPipeCreatorDataSource(
        ...     datapipe_creators={
        ...         "train": {
        ...             '_target_': "gravitorch.creators.datapipe.SequentialIterDataPipeCreator",
        ...             "config": [
        ...                 {
        ...                     OBJECT_TARGET: "gravitorch.data.datapipes.iter.SourceWrapperIterDataPipe",
        ...                    "data": [1, 2, 3, 4],
        ...                },
        ...             ],
        ...         },
        ...         "val": {
        ...             '_target_': "gravitorch.creators.datapipe.SequentialIterDataPipeCreator",
        ...             "config": [
        ...                 {
        ...                     '_target_': "gravitorch.data.datapipes.iter.SourceWrapperIterDataPipe",
        ...                     "data": ["a", "b", "c"],
        ...                 },
        ...             ],
        ...         },
        ...     }
        ... )
        # Note that both examples lead to the same result.
    """

    def __init__(self, datapipe_creators: dict[str, Union[BaseIterDataPipeCreator, dict]]):
        self._asset_manager = AssetManager()
        logger.info("Initializing the IterDataPipe creators...")
        self._datapipe_creators = {
            key: setup_iter_datapipe_creator(creator) for key, creator in datapipe_creators.items()
        }
        logger.info(f"IterDataPipe creators:\n{to_torch_mapping_str(self._datapipe_creators)}")

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  {str_add_indent(to_torch_mapping_str(self._datapipe_creators))}\n)"
        )

    def attach(self, engine: BaseEngine) -> None:
        r"""Attaches the current data source to the provided engine.

        This method can be used to set up events or logs some stats
        to the engine.

        Args:
            engine (``BaseEngine``): Specifies the engine.

        Example:

        .. code-block:: python

            >>> from gravitorch.datasources import IterDataPipeCreatorDataSource
            >>> data_source = IterDataPipeCreatorDataSource(datapipe_creators={...})
            >>> from gravitorch.engines import AlphaEngine
            >>> my_engine = AlphaEngine()  # Work with any engine
            >>> data_source.attach(my_engine)
        """
        logger.info("Attach the data source to an engine")

    def get_asset(self, asset_id: str) -> Any:
        r"""Gets a data asset from this data source.

        This method is useful to access some data variables/parameters
        that are not available before to load/preprocess the data.

        Args:
            asset_id (str): Specifies the ID of the asset.

        Returns:
            The asset.

        Raises:
            ``AssetNotFoundError`` if you try to access an asset
                that does not exist.

        Example:

        .. code-block:: python

            >>> from gravitorch.datasources import IterDataPipeCreatorDataSource
            >>> data_source = IterDataPipeCreatorDataSource(datapipe_creators={...})
            >>> my_asset = data_source.get_asset('my_asset_id')
        """
        return self._asset_manager.get_asset(asset_id)

    def has_asset(self, asset_id: str) -> bool:
        r"""Indicates if the asset exists or not.

        Args:
            asset_id (str): Specifies the ID of the asset.

        Returns:
            bool: ``True`` if the asset exists, otherwise ``False``.

        Example:

        .. code-block:: python

            >>> from gravitorch.datasources import IterDataPipeCreatorDataSource
            >>> data_source = IterDataPipeCreatorDataSource(datapipe_creators={...})
            >>> data_source.has_asset('my_asset_id')
            False
        """
        return self._asset_manager.has_asset(asset_id)

    def get_data_loader(self, loader_id: str, engine: Optional[BaseEngine] = None) -> Iterable[T]:
        r"""Gets a data loader.

        Args:
            loader_id (str): Specifies the ID of the data loader to
                get.
            engine (``BaseEngine`` or ``None``, optional): Specifies
                an engine. The engine can be used to initialize the
                data loader by using the current epoch value.
                Default: ``None``

        Returns:
            ``Iterable``: A data loader.

        Raises:
            ``LoaderNotFoundError`` if the loader does not exist.

        Example:

        .. code-block:: python

            >>> from gravitorch.datasources import IterDataPipeCreatorDataSource
            >>> data_source = IterDataPipeCreatorDataSource(datapipe_creators={...})
            # Get the data loader associated to the ID 'train'
            >>> data_loader = data_source.get_data_loader('train')
            # Get a data loader that can use information from an engine
            >>> from gravitorch.engines import AlphaEngine
            >>> my_engine = AlphaEngine()  # Work with any engine
            >>> data_loader = data_source.get_data_loader('train', my_engine)
        """
        if not self.has_data_loader(loader_id):
            raise LoaderNotFoundError(f"{loader_id} does not exist")
        return self._create_datapipe(loader_id=loader_id, engine=engine)

    def has_data_loader(self, loader_id: str) -> bool:
        r"""Indicates if the data source has a data loader with the given ID.

        Args:
            loader_id (str): Specifies the ID of the data loader.

        Returns:
            bool: ``True`` if the data loader exists, ``False``
                otherwise.

        Example:

        .. code-block:: python

            >>> from gravitorch.datasources import IterDataPipeCreatorDataSource
            >>> data_source = IterDataPipeCreatorDataSource(datapipe_creators={...})
            # Check if the data source has a data loader for ID 'train'
            >>> data_source.has_data_loader('train')
            True or False
            # Check if the data source has a data loader for ID 'eval'
            >>> data_source.has_data_loader('eval')
            True or False
        """
        return loader_id in self._datapipe_creators

    def _create_datapipe(
        self, loader_id: str, engine: Optional[BaseEngine] = None
    ) -> IterDataPipe[T]:
        r"""Creates an ``IterDataPipe`` object.

        Args:
            loader_id (str): Specifies the ID of the data loader to
                get.
            engine (``BaseEngine`` or ``None``, optional): Specifies
                an engine. The engine can be used to initialize the
                data loader by using the current epoch value.
                Default: ``None``

        Returns:
            ``IterDataPipe``: An ``IterDataPipe`` object.
        """
        return self._datapipe_creators[loader_id].create(engine=engine)


class DataCreatorIterDataPipeCreatorDataSource(IterDataPipeCreatorDataSource):
    r"""Implements a data source that creates data loaders using
    ``IterDataPipe`` creators.

    Unlike ``IterDataPipeCreatorDataSource``, each ``IterDataPipe``
    creator takes as input (``source_inputs``) the data created by a
    ``BaseDataCreator`` object if it is defined. If no
    ``BaseDataCreator`` object is defined, ``source_inputs`` of the
    ``IterDataPipe`` creator is set to ``None``.

    Args:
        datapipe_creators (dict): Specifies the ``IterDataPipe``
            creators or their configurations. Each key is associated
            to a loader ID. For example if you want to use a
            ``'train'`` data loader, you need to map this key to a
            ``BaseIterDataPipeCreator`` object or its configuration.
            Each ``BaseIterDataPipeCreator`` object contains the
            recipe to create an ``IterDataPipe`` object.
        data_creators (dict): Specifies the data creators or their
            configurations. Each key is associated to a loader ID.
            For example if you want to create data for the ``'train'``
            loader, you need to map this key to a ``BaseDataCreator``
            object or its configuration.
    """

    def __init__(
        self,
        datapipe_creators: dict[str, Union[BaseIterDataPipeCreator, dict]],
        data_creators: dict[str, Union[BaseDataCreator, dict]],
    ):
        super().__init__(datapipe_creators)
        logger.info("Initializing the data creators...")
        self._data_creators = {
            key: setup_data_creator(creator) for key, creator in data_creators.items()
        }
        logger.info(f"Data creators:\n{to_torch_mapping_str(self._data_creators)}")
        logger.info("Creating data...")
        self._data = {key: creator.create() for key, creator in self._data_creators.items()}
        logger.info(f"Data:\n{concise_summary(self._data)}")

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            "  data_creators\n"
            f"    {str_add_indent(to_torch_mapping_str(self._data_creators), num_spaces=4)}\n"
            "  datapipe_creators\n"
            f"    {str_add_indent(to_torch_mapping_str(self._datapipe_creators), num_spaces=4)}"
            "\n)"
        )

    def _create_datapipe(self, loader_id: str, engine: Optional[BaseEngine] = None) -> IterDataPipe:
        r"""Creates an ``IterDataPipe`` object.

        Args:
            loader_id (str): Specifies the ID of the data loader to
                get.
            engine (``BaseEngine`` or ``None``, optional): Specifies
                an engine. The engine can be used to initialize the
                data loader by using the current epoch value.
                Default: ``None``

        Returns:
            ``IterDataPipe``: An ``IterDataPipe`` object.
        """
        source_input = self._data.get(loader_id, None)
        return self._datapipe_creators[loader_id].create(
            engine=engine,
            source_inputs=source_input if source_input is None else (source_input,),
        )



__all__ = ["VanillaLRSchedulerCreator"]

import logging
from typing import Optional, Union

from torch.optim import Optimizer

from gravitorch import constants as ct
from gravitorch.creators.lr_scheduler.base import BaseLRSchedulerCreator
from gravitorch.engines.base import BaseEngine
from gravitorch.handlers.base import BaseHandler
from gravitorch.handlers.utils import setup_handler
from gravitorch.lr_schedulers.base import LRSchedulerType, setup_lr_scheduler
from gravitorch.utils.format import str_add_indent

logger = logging.getLogger(__name__)


class VanillaLRSchedulerCreator(BaseLRSchedulerCreator):
    r"""Implements a vanilla a learning rate (LR) scheduler creator.

    This LR scheduler creator has two main inputs: an input to
    configure the LR scheduler and one to manage the LR scheduler.
    The LR scheduler manager is responsible to create events to
    control the LR scheduler.

    Args:
        lr_scheduler_config (dict or ``None``): Specifies the LR
            scheduler configuration. If ``None``, no LR scheduler
            is created and ``None`` will be returned by the ``create``
            method. Default: ``None``
        lr_scheduler_handler (``BaseLRSchedulerManager`` or dict or
            ``None``): Specifies the LR scheduler handler. The LR
            scheduler manager is used only if the LR scheduler can
            be created. If ``None``, no LR scheduler manager is
            created and the user is responsible to manage the LR
            scheduler. Default: ``None``
        add_module_to_engine (bool, optional): If ``True``, the LR
            scheduler is added to the engine state, so the LR
            scheduler state is stored when the engine creates a
            checkpoint. Default: ``True``
    """

    def __init__(
        self,
        lr_scheduler_config: Union[dict, None] = None,
        lr_scheduler_handler: Union[BaseHandler, dict, None] = None,
        add_module_to_engine: bool = True,
    ):
        self._lr_scheduler_config = lr_scheduler_config
        self._lr_scheduler_manager = setup_handler(lr_scheduler_handler)
        logger.info(f"lr_scheduler_handler:\n{lr_scheduler_handler}")
        self._add_module_to_engine = bool(add_module_to_engine)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  lr_scheduler_handler={str_add_indent(str(self._lr_scheduler_manager))},\n"
            f"  add_module_to_engine={self._add_module_to_engine},\n"
            ")"
        )

    def create(
        self, engine: BaseEngine, optimizer: Optional[Optimizer]
    ) -> Optional[LRSchedulerType]:
        lr_scheduler = setup_lr_scheduler(
            optimizer=optimizer, lr_scheduler=self._lr_scheduler_config
        )
        if lr_scheduler is None:
            return None

        logger.info(f"lr_scheduler:\n{lr_scheduler}")
        if self._add_module_to_engine:
            logger.info(f"Adding a LR scheduler to the engine state (key: {ct.LR_SCHEDULER})...")
            engine.add_module(ct.LR_SCHEDULER, lr_scheduler)

        if self._lr_scheduler_manager:
            logger.info("Attaching a LR scheduler manager to the engine...")
            self._lr_scheduler_manager.attach(engine=engine)
        else:
            logger.warning(
                "No LR scheduler manager is set. If you do not use a LR scheduler manager, you "
                "need to manage 'manually' the LR scheduler"
            )
        return lr_scheduler



r"""This module implements some helper functions to log information."""

__all__ = ["log_run_info"]

import logging
from pathlib import Path

from omegaconf import DictConfig, OmegaConf

import gravitorch
from gravitorch.utils.path import get_original_cwd

logger = logging.getLogger(__name__)

# ASCII logo generated from https://ascii.co.uk/art/cats
GRAVITORCH_LOGO = rf"""

        /\_/\  /\          /\_/\  /\          /\_/\  /\          /\_/\  /\
       / o o \ \ \        / o o \ \ \        / o o \ \ \        / o o \ \ \
      /   Y   \/ /       /   Y   \/ /       /   Y   \/ /       /   Y   \/ /
     /         \/       /         \/       /         \/       /         \/
     \ | | | | /        \ | | | | /        \ | | | | /        \ | | | | /
      `|_|-|_|'          `|_|-|_|'          `|_|-|_|'          `|_|-|_|'

version: {gravitorch.__version__}"""


def log_run_info(config: DictConfig) -> None:
    """Log some information about the current run.

    Args:
        config (``omegaconf.DictConfig``): Specifies the config of
            the run.
    """
    logger.info(GRAVITORCH_LOGO)
    logger.info("Original working directory: %s", get_original_cwd())
    logger.info("Current working directory: %s", Path.cwd())
    logger.info("Config:\n%s", OmegaConf.to_yaml(config))



r"""This module defines the base class for the evaluation loops."""

__all__ = ["BaseEvaluationLoop", "BaseBasicEvaluationLoop"]

import logging
import sys
from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import Any, Union

from objectory import AbstractFactory
from torch.nn import Module

from gravitorch import constants as ct
from gravitorch.distributed import comm as dist
from gravitorch.engines.base import BaseEngine
from gravitorch.engines.events import EngineEvents
from gravitorch.utils.cudamem import log_max_cuda_memory_allocated
from gravitorch.utils.evaluation_loops.conditions import (
    BaseEvalCondition,
    EveryEpochEvalCondition,
)
from gravitorch.utils.history import MinScalarHistory
from gravitorch.utils.loop_observers import BaseLoopObserver, setup_loop_observer
from gravitorch.utils.metric_tracker import ScalarMetricTracker
from gravitorch.utils.profilers import BaseProfiler, setup_profiler
from gravitorch.utils.seed import manual_seed
from gravitorch.utils.timing import BatchLoadingTimer

logger = logging.getLogger(__name__)


class BaseEvaluationLoop(ABC, metaclass=AbstractFactory):
    r"""Defines the evaluation loop base class.

    To implement your own evaluation loop, you will need to define
    the ``eval`` method. If the evaluation loop has state values,
    you also need to implement the following methods:

        - ``load_state_dict``
        - ``state_dict``
    """

    @abstractmethod
    def eval(self, engine: BaseEngine) -> None:
        r"""Evaluates the model on the evaluation dataset.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """

    def load_state_dict(self, state_dict: dict[str, Any]) -> None:
        r"""Sets up the evaluation loop from a dictionary containing the state
        values.

        Args:
            state_dict (dict): Specifies a dictionary
                containing state keys with values.
        """

    def state_dict(self) -> dict[str, Any]:
        r"""Returns a dictionary containing state values.

        Returns:
            dict: The state values in a dict.
        """
        return {}


class BaseBasicEvaluationLoop(BaseEvaluationLoop):
    r"""Implements a simple evaluation loop to evaluate a model on a given
    dataset.

    Args:
        tag (str, optional): Specifies the tag which is used to log
            metrics. Default: ``"eval"``
        condition (``BaseEvalCondition`` or dict or None): Specifies
            the condition to evaluate the loop or its configuration.
            If ``None``, the ``EveryEpochEvalCondition(every=1)`` is
            used.  Default ``None``
        observer (``BaseLoopObserver`` or dict or None, optional):
            Specifies the loop observer or its configuration.
            If ``None``, the ``NoOpLoopObserver`` is instantiated.
            Default: ``None``
        profiler (``BaseProfiler`` or dict or None, optional):
            Specifies the profiler or its configuration. If ``None``,
            the ``NoOpProfiler`` is instantiated. Default: ``None``
    """

    def __init__(
        self,
        tag: str = "eval",
        condition: Union[BaseEvalCondition, dict, None] = None,
        observer: Union[BaseLoopObserver, dict, None] = None,
        profiler: Union[BaseProfiler, dict, None] = None,
    ):
        self._tag = str(tag)
        self._condition = self._setup_condition(condition)
        logger.info(f"condition:\n{self._condition}")
        self._observer = setup_loop_observer(observer)
        logger.info(f"observer:\n{self._observer}")
        self._profiler = setup_profiler(profiler)
        logger.info(f"profiler:\n{self._profiler}")

    def eval(self, engine: BaseEngine) -> None:
        r"""Evaluates the model on the evaluation dataset.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        dist.barrier()
        if not engine.data_source.has_data_loader(self._tag) or not self._condition(engine):
            return
        logger.info(f"Evaluating model for epoch {engine.epoch}")

        self._prepare_evaluation(engine)
        engine.fire_event(EngineEvents.EVAL_EPOCH_STARTED)

        model, data_loader = self._prepare_model_data_loader(engine)

        # Evaluate the model on each mini-match in the dataset.
        metrics = ScalarMetricTracker()
        data_loader = BatchLoadingTimer(data_loader, epoch=engine.epoch, prefix=f"{self._tag}/")
        self._observer.start(engine)
        dist.barrier()

        with self._profiler as profiler:
            for batch in data_loader:
                # Run forward on the given batch.
                output = self._eval_one_batch(engine, model, batch)
                metrics.update(output)
                self._observer.update(engine=engine, model_input=batch, model_output=output)
                profiler.step()

        # To be sure the progress bar is displayed before the following lines
        sys.stdout.flush()
        dist.barrier()
        self._observer.end(engine)

        # Log some evaluation metrics to the engine.
        data_loader.log_stats(engine=engine)
        metrics.log_average_value(engine=engine, prefix=f"{self._tag}/")
        log_max_cuda_memory_allocated()
        dist.barrier()

        engine.fire_event(EngineEvents.EVAL_EPOCH_COMPLETED)

    def _prepare_evaluation(self, engine: BaseEngine) -> None:
        r"""Prepares the evaluation.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        # Fix the random seed for reproducibility purpose.
        manual_seed(engine.random_seed + engine.epoch + engine.max_epochs * dist.get_rank())
        engine.model.eval()

        if not engine.has_history(f"{self._tag}/{ct.LOSS}"):
            engine.add_history(MinScalarHistory(f"{self._tag}/{ct.LOSS}"))

    def _setup_condition(
        self, condition: Union[BaseEvalCondition, dict, None]
    ) -> BaseEvalCondition:
        r"""Sets up the condition.

        The condition is instantiated from its configuration by using
        the ``BaseEvalCondition`` factory function.

        Args:
            condition (``BaseEvalCondition`` or dict or None):
                Specifies the condition or its configuration.
                If ``None``, the ``EveryEpochEvalCondition(every=1)``
                is instantiated.

        Returns:
            ``BaseEvalCondition``: The state.
        """
        condition = condition or EveryEpochEvalCondition(every=1)
        if isinstance(condition, dict):
            logger.info("Initializing the condition from its configuration...")
            condition = BaseEvalCondition.factory(**condition)

        logger.info(f"condition:\n{condition}")
        return condition

    @abstractmethod
    def _eval_one_batch(self, engine: BaseEngine, model: Module, batch: Any) -> dict:
        r"""Evaluates the model on the given batch.

        Args:
            engine (``BaseEngine``): Specifies the engine.
            model (``torch.nn.Module``): Specifies the model to evaluate.
            batch: Specifies the batch of data.

        Returns:
            dict: Some results about the batch.
        """

    @abstractmethod
    def _prepare_model_data_loader(self, engine: BaseEngine) -> tuple[Module, Iterable]:
        r"""Prepares the model, optimizer and data loader.

        Args:
            engine (``BaseEngine``): Specifies the engine.

        Returns:
            ``torch.nn.Module``, ``Iterable``: A tuple with the model
                and the data loader.
        """



r"""This module implements some utility functions for the data summary."""

__all__ = ["setup_data_summary"]

import logging
from typing import Union

from gravitorch.utils.data_summary.base import BaseDataSummary
from gravitorch.utils.data_summary.noop import NoOpDataSummary
from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


def setup_data_summary(data_summary: Union[BaseDataSummary, dict, None]) -> BaseDataSummary:
    r"""Sets up a data summary object.

    The data summary module is instantiated from its configuration by
    using the ``BaseDataSummary`` factory function.

    Args:
        data_summary (``BaseDataSummary`` or dict or ``None``):
            Specifies the data summary object or its configuration.
            If ``None``, a ``NoOpDataSummary`` object is returned.

    Returns:
        ``BaseDataSummary``: The data summary object.
    """
    if data_summary is None:
        return NoOpDataSummary()
    if isinstance(data_summary, dict):
        logger.info(
            "Initializing the data summary module from its configuration... "
            f"{str_target_object(data_summary)}"
        )
        data_summary = BaseDataSummary.factory(**data_summary)
    return data_summary



__all__ = ["AverageMeter"]

from collections.abc import Iterable
from typing import Any

from gravitorch.distributed.ddp import SUM, sync_reduce
from gravitorch.utils.format import to_pretty_dict_str
from gravitorch.utils.meters.exceptions import EmptyMeterError


class AverageMeter:
    r"""Defines a class to compute and store the average value of float number.

    Args:
        total (float, optional): Specifies the initial total value.
            Default: ``0.0``
        count (int, optional): Specifies the initial count value.
            Default: ``0``

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.meters import AverageMeter
        >>> meter = AverageMeter()
        >>> for i in range(11):
        ...     meter.update(i)
        >>> meter.average()
        5.0
    """

    def __init__(self, total: float = 0.0, count: int = 0):
        self._total = float(total)
        self._count = int(count)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(count={self._count:,}, total={self._total})"

    def __str__(self) -> str:
        stats = to_pretty_dict_str(
            {
                "average": self.average() if self.count else "N/A (empty)",
                "count": self.count,
                "total": self.total,
            },
            indent=2,
        )
        return f"{self.__class__.__qualname__}\n{stats}"

    @property
    def count(self) -> int:
        r"""int: The number of examples in the meter since the last
        reset."""
        return self._count

    @property
    def total(self) -> float:
        r"""float: The total of the values added to the meter since the
        last reset."""
        return self._total

    def all_reduce(self) -> "AverageMeter":
        r"""Reduces the meter values across all machines in such a way that all
        get the final result.

        The total value is reduced by summing all the sum values
        (1 total value per distributed process).
        The count value is reduced by summing all the count values
        (1 count value per distributed process).

        Returns:
            ``AverageMeter``: The reduced meter.

        Example usage:

        .. code-block:: python

            >>> from gravitorch.utils.meters import AverageMeter
            >>> meter = AverageMeter()
            >>> meter.update(6)
            >>> reduced_meter = meter.all_reduce()
        """
        return AverageMeter(
            total=sync_reduce(self._total, SUM),
            count=sync_reduce(self._count, SUM),
        )

    def average(self) -> float:
        r"""Computes the average value.

        Returns:
            float: The average value.

        Raises:
            ``EmptyMeterError`` if the meter is empty.
        """
        if not self._count:
            raise EmptyMeterError("The average meter is empty")
        return self._total / float(self._count)

    def clone(self) -> "AverageMeter":
        r"""Creates a copy of the current meter.

        Returns:
            ``AverageMeter``: A copy of the current meter.
        """
        return AverageMeter(total=self.total, count=self.count)

    def equal(self, other) -> bool:
        r"""Indicates if two meters are equal or not.

        Args:
            other: Specifies the value to compare.

        Returns:
            bool: ``True`` if the meters are equal,
                ``False`` otherwise.
        """
        if not isinstance(other, AverageMeter):
            return False
        return self.state_dict() == other.state_dict()

    def merge(self, meters: Iterable["AverageMeter"]) -> "AverageMeter":
        r"""Merges several meters with the current meter and returns a new
        meter.

        Args:
            meters (iterable): Specifies the meters to merge to the
                current meter.

        Returns:
            ``AverageMeter``: The merged meter.
        """
        count, total = self.count, self.total
        for meter in meters:
            count += meter.count
            total += meter.total
        return AverageMeter(total=total, count=count)

    def merge_(self, meters: Iterable["AverageMeter"]) -> None:
        r"""Merges several meters into the current meter.

        In-place version of ``merge``.

        Args:
            meters (iterable): Specifies the meters to merge to the
                current meter.
        """
        for meter in meters:
            self._count += meter.count
            self._total += meter.total

    def load_state_dict(self, state_dict: dict[str, Any]) -> None:
        r"""Loads a state to the history tracker.

        Args:
            state_dict (dict): Specifies a dictionary containing state
                keys with values.
        """
        self._total = state_dict["total"]
        self._count = state_dict["count"]

    def reset(self) -> None:
        r"""Reset the meter."""
        self._total = 0.0
        self._count = 0

    def state_dict(self) -> dict[str, Any]:
        r"""Returns a dictionary containing state values.

        Returns:
            dict: The state values in a dict.
        """
        return {"count": self._count, "total": self._total}

    def sum(self) -> float:
        r"""Computes the sum value.

        Returns:
            float: The sum value.

        Raises:
            ``EmptyMeterError`` if the meter is empty.
        """
        if not self._count:
            raise EmptyMeterError("The average meter is empty")
        return self._total

    def update(self, value: float, num_examples: int = 1) -> None:
        r"""Updates the meter given a new value and the number of examples.

        Args:
            value (float): Specifies the value to add to the meter.
            num_examples (int, optional): Specifies the number of
                examples. This argument is mainly used to deal with
                mini-batches of different sizes. Default: ``1``
        """
        self._total += float(value) * num_examples
        self._count += num_examples



__all__ = ["NoOpOptimizer"]

from collections.abc import Iterable
from typing import Union

from torch import Tensor
from torch.optim import Optimizer


class NoOpOptimizer(Optimizer):
    r"""Implements a no-op optimizer.

    This optimizer cannot be used to train a model. This optimizer can
    be used to simulate an optimizer that does not update the model
    parameters.

    Args:
        params: This input is not used. It is here to make it
            compatible with the other optimizers.
    """

    def __init__(self, params: Union[Iterable[Tensor], Iterable[dict]]):
        r"""Do nothing."""

    def load_state_dict(self, state_dict: dict) -> None:
        r"""Do nothing."""

    def state_dict(self) -> dict:
        return {}

    def step(self, closure=None):
        r"""Do nothing."""



__all__ = ["PyTorchBatchSaver"]

import logging
from pathlib import Path
from typing import Any, Union

from gravitorch.engines.base import BaseEngine
from gravitorch.utils.io import save_pytorch
from gravitorch.utils.loop_observers.base import BaseLoopObserver
from gravitorch.utils.path import sanitize_path

logger = logging.getLogger(__name__)


class PyTorchBatchSaver(BaseLoopObserver):
    r"""Implements a model batch saver to store batches in a PyTorch file
    (``torch.save``).

    Args:
        path (``pathlib.Path`` or str): Specifies the path where to
            store the examples.
        max_num_batches (int, optional): Specifies the maximum number
            of batches to save. Default: ``1000``
    """

    def __init__(self, path: Union[Path, str], max_num_batches: int = 1000):
        self._path = sanitize_path(path)
        self._max_num_batches = max_num_batches
        self._batches = []

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(path={self._path}, "
            f"max_num_batches={self._max_num_batches:,})"
        )

    def update(self, engine: BaseEngine, model_input: Any, model_output: Any) -> None:
        r"""Update the observer.

        Args:
            engine (``BaseEngine``): Specifies the engine.
            model_input: Specifies a batch of model input.
            model_output: Specifies a batch of model output.
        """
        if len(self._batches) < self._max_num_batches:
            self._batches.append({"model_input": model_input, "model_output": model_output})

    def start(self, engine: BaseEngine) -> None:
        r"""Resets the model batch saver.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        self._batches.clear()

    def end(self, engine: BaseEngine) -> None:
        r"""Saves the batches in a PyTorch file.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        logger.info(f"Saving batches in {self._path}")
        save_pytorch(self._batches, self._path)



__all__ = [
    "BasePartitioner",
    "DDPPartitioner",
    "EpochShufflePartitioner",
    "EvenPartitioner",
    "FixedSizePartitioner",
    "SequentialPartitioner",
    "SyncParallelPartitioner",
    "TrivialPartitioner",
    "setup_partitioner",
]

from gravitorch.data.partitioners.base import BasePartitioner, setup_partitioner
from gravitorch.data.partitioners.distributed import (
    DDPPartitioner,
    SyncParallelPartitioner,
)
from gravitorch.data.partitioners.even import EvenPartitioner
from gravitorch.data.partitioners.fixed_size import FixedSizePartitioner
from gravitorch.data.partitioners.sequential import SequentialPartitioner
from gravitorch.data.partitioners.shuffling import EpochShufflePartitioner
from gravitorch.data.partitioners.trivial import TrivialPartitioner



r"""This package contains the implementation of some data summaries."""

__all__ = [
    "BaseContinuousDataSummary",
    "BaseDataSummary",
    "BaseDiscreteDataSummary",
    "BaseSequenceDataSummary",
    "EmptyDataSummaryError",
    "FloatDataSummary",
    "FloatTensorDataSummary",
    "FloatTensorSequenceDataSummary",
    "IntegerDataSummary",
    "LongTensorDataSummary",
    "LongTensorSequenceDataSummary",
    "NoOpDataSummary",
    "setup_data_summary",
]

from gravitorch.utils.data_summary.base import BaseDataSummary, EmptyDataSummaryError
from gravitorch.utils.data_summary.continuous import BaseContinuousDataSummary
from gravitorch.utils.data_summary.discrete import BaseDiscreteDataSummary
from gravitorch.utils.data_summary.float_tensor import (
    FloatTensorDataSummary,
    FloatTensorSequenceDataSummary,
)
from gravitorch.utils.data_summary.float_value import FloatDataSummary
from gravitorch.utils.data_summary.integer import IntegerDataSummary
from gravitorch.utils.data_summary.long_tensor import (
    LongTensorDataSummary,
    LongTensorSequenceDataSummary,
)
from gravitorch.utils.data_summary.noop import NoOpDataSummary
from gravitorch.utils.data_summary.sequence import BaseSequenceDataSummary
from gravitorch.utils.data_summary.utils import setup_data_summary



__all__ = [
    "cpu_human_summary",
    "log_system_info",
    "swap_memory_human_summary",
    "virtual_memory_human_summary",
]

import logging

import psutil

from gravitorch.utils.format import to_human_readable_byte_size

logger = logging.getLogger(__name__)


def cpu_human_summary() -> str:
    r"""Gets a human-readable summary of the CPU usage.

    Returns:
        str: The human-readable summary

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.sysinfo import cpu_human_summary
        >>> cpu_human_summary()
        CPU - logical/physical count: 4/2 | percent: 42.0 | load 1/5/15min: 42.42/36.48/32.68 %
    """
    loadavg = tuple(100.0 * x / psutil.cpu_count() for x in psutil.getloadavg())
    return (
        f"CPU - logical/physical count: {psutil.cpu_count()}/{psutil.cpu_count(logical=False)} | "
        f"percent: {psutil.cpu_percent()} % | "
        f"load 1/5/15min: {loadavg[0]:.2f}/{loadavg[1]:.2f}/{loadavg[2]:.2f} %"
    )


def log_system_info() -> None:
    r"""Log information about the system.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.sysinfo import log_system_info
        >>> log_system_info()
    """
    logger.info(cpu_human_summary())
    logger.info(virtual_memory_human_summary())
    logger.info(swap_memory_human_summary())


def swap_memory_human_summary() -> str:
    r"""Gets a human-readable summary of the swap memory usage.

    Returns:
        str: The human-readable summary

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.sysinfo import swap_memory_human_summary
        >>> swap_memory_human_summary()
        swap memory - total: 17.00 GB | used: 15.66 GB | free: 1.34 GB | percent: 92.1% | sin: 835.39 GB | sout: 45.64 GB  # noqa: E501,B950
    """
    swap = psutil.swap_memory()
    return (
        f"swap memory - total: {to_human_readable_byte_size(swap.total)} | "
        f"used: {to_human_readable_byte_size(swap.used)} | "
        f"free: {to_human_readable_byte_size(swap.free)} | "
        f"percent: {swap.percent} % | "
        f"sin: {to_human_readable_byte_size(swap.sin)} | "
        f"sout: {to_human_readable_byte_size(swap.sout)}"
    )


def virtual_memory_human_summary() -> str:
    r"""Gets a human-readable summary of the virtual memory usage.

    Returns:
        str: The human-readable summary

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.sysinfo import virtual_memory_human_summary
        >>> virtual_memory_human_summary()
        virtual memory - total: 16.00 GB | available: 2.89 GB | percent: 81.9% | used: 5.43 GB | free: 28.14 MB  # noqa: E501,B950
    """
    vm = psutil.virtual_memory()
    return (
        f"virtual memory - total: {to_human_readable_byte_size(vm.total)} | "
        f"available: {to_human_readable_byte_size(vm.available)} | "
        f"percent: {vm.percent} % | "
        f"used: {to_human_readable_byte_size(vm.used)} | "
        f"free: {to_human_readable_byte_size(vm.free)}"
    )



__all__ = [
    "BaseKaimingParameterInitializer",
    "KaimingNormalParameterInitializer",
    "KaimingUniformParameterInitializer",
    "recursive_kaiming_normal_",
    "recursive_kaiming_uniform_",
]

import logging

from torch import nn

from gravitorch.engines.base import BaseEngine
from gravitorch.utils.parameter_initializers.base import BaseDefaultParameterInitializer

logger = logging.getLogger(__name__)


class BaseKaimingParameterInitializer(BaseDefaultParameterInitializer):
    r"""Implements a model parameter initializer with the Kaiming Normal or
    uniform strategy.

    Args:
        neg_slope (float, optional): Specifies the negative slope of
            the rectifier used after this layer (only used with
            ``'leaky_relu'``). Default: ``0.0``
        mode (str, optional): either ``'fan_in'`` or ``'fan_out'``.
            Choosing ``'fan_in'`` preserves the magnitude of the
            variance of the weights in the forward pass. Choosing
            ``'fan_out'`` preserves the magnitudes in the backwards
            pass. Default: ``'fan_in'``
        nonlinearity (str, optional): the non-linear function
            (`nn.functional` name), recommended to use only with
            ``'relu'`` or ``'leaky_relu'``. Default: ``'leaky_relu'``
        learnable_only (bool, optional): If ``True``, only the
            learnable parameters are initialized, otherwise all the
            parameters are initialized. Default: ``True``
        show_stats (bool, optional): If ``True``, the parameter
            statistics are shown at the end of the initialization.
            Default: ``True``
    """

    def __init__(
        self,
        neg_slope: float = 0.0,
        mode: str = "fan_in",
        nonlinearity: str = "leaky_relu",
        learnable_only: bool = True,
        show_stats: bool = True,
    ):
        super().__init__(show_stats=show_stats)
        self._neg_slope = float(neg_slope)
        self._mode = str(mode)
        self._nonlinearity = str(nonlinearity)
        self._learnable_only = bool(learnable_only)

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(neg_slope={self._neg_slope}, mode={self._mode}, "
            f"nonlinearity={self._nonlinearity}, learnable_only={self._learnable_only})"
        )


class KaimingNormalParameterInitializer(BaseKaimingParameterInitializer):
    r"""Implements a model parameter initializer with the Kaiming Normal
    strategy."""

    def _initialize(self, engine: BaseEngine) -> None:
        logger.info("Initializing model parameters with the Kaiming Normal strategy...")
        recursive_kaiming_normal_(
            module=engine.model,
            neg_slope=self._neg_slope,
            mode=self._mode,
            nonlinearity=self._nonlinearity,
            learnable_only=self._learnable_only,
        )


class KaimingUniformParameterInitializer(BaseKaimingParameterInitializer):
    r"""Implements a model parameter initializer with the Kaiming uniform
    strategy."""

    def _initialize(self, engine: BaseEngine) -> None:
        logger.info("Initializing model parameters with the Kaiming uniform strategy...")
        recursive_kaiming_uniform_(
            module=engine.model,
            neg_slope=self._neg_slope,
            mode=self._mode,
            nonlinearity=self._nonlinearity,
            learnable_only=self._learnable_only,
        )


def recursive_kaiming_normal_(
    module: nn.Module,
    neg_slope: float = 0.0,
    mode: str = "fan_in",
    nonlinearity: str = "leaky_relu",
    learnable_only: bool = True,
) -> None:
    r"""Initialize the parameters of the module with the Kaiming Normal
    initialization.

    Args:
        module (``torch.nn.Module``): Specifies the module to
            initialize.
        neg_slope (float, optional): Specifies the negative slope of
            the rectifier used after this layer (only used with
            ``'leaky_relu'``). Default: ``0.0``
        mode (str, optional): either ``'fan_in'`` or ``'fan_out'``.
            Choosing ``'fan_in'`` preserves the magnitude of the
            variance of the weights in the forward pass. Choosing
            ``'fan_out'`` preserves the magnitudes in the backwards
            pass. Default: ``'fan_in'``
        nonlinearity (str, optional): the non-linear function
            (`nn.functional` name), recommended to use only with
            ``'relu'`` or ``'leaky_relu'``. Default: ``'leaky_relu'``
        learnable_only (bool, optional): If ``True``, only the
            learnable parameters are initialized, otherwise all the
            parameters are initialized. Default: ``True``

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.parameter_initializers import recursive_kaiming_normal_
        >>> from torch import nn
        >>> net = nn.Sequential(nn.Linear(4, 6), nn.ReLU(), nn.BatchNorm1d(6), nn.Linear(6, 1))
        >>> recursive_kaiming_normal_(net)
    """
    for params in module.parameters():
        if params.ndim > 1 and (not learnable_only or learnable_only and params.requires_grad):
            nn.init.kaiming_normal_(params.data, a=neg_slope, mode=mode, nonlinearity=nonlinearity)


def recursive_kaiming_uniform_(
    module: nn.Module,
    neg_slope: float = 0.0,
    mode: str = "fan_in",
    nonlinearity: str = "leaky_relu",
    learnable_only: bool = True,
) -> None:
    r"""Initialize the parameters of the module with the Kaiming uniform
    initialization.

    Args:
        module (``torch.nn.Module``): Specifies the module to
            initialize.
        neg_slope (float, optional): Specifies the negative slope of
            the rectifier used after this layer (only used with
            ``'leaky_relu'``). Default: ``0.0``
        mode (str, optional): either ``'fan_in'`` or ``'fan_out'``.
            Choosing ``'fan_in'`` preserves the magnitude of the
            variance of the weights in the forward pass. Choosing
            ``'fan_out'`` preserves the magnitudes in the backwards
            pass. Default: ``'fan_in'``
        nonlinearity (str, optional): the non-linear function
            (`nn.functional` name), recommended to use only with
            ``'relu'`` or ``'leaky_relu'``. Default: ``'leaky_relu'``
        learnable_only (bool, optional): If ``True``, only the
            learnable parameters are initialized, otherwise all
            the parameters are initialized. Default: ``True``

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.parameter_initializers import recursive_kaiming_uniform_
        >>> from torch import nn
        >>> net = nn.Sequential(nn.Linear(4, 6), nn.ReLU(), nn.BatchNorm1d(6), nn.Linear(6, 1))
        >>> recursive_kaiming_uniform_(net)
    """
    for params in module.parameters():
        if params.ndim > 1 and (not learnable_only or learnable_only and params.requires_grad):
            nn.init.kaiming_uniform_(params.data, a=neg_slope, mode=mode, nonlinearity=nonlinearity)



r"""This module implements some utilities for the loggers."""

import logging
from collections.abc import Generator
from contextlib import contextmanager
from typing import Union


@contextmanager
def disable_logging(level: Union[int, str] = logging.CRITICAL) -> Generator[None, None, None]:
    r"""Context manager to temporarily disable the logging.

    All logging calls of severity ``level`` and below will be
    disabled.

    Args:
        level (int or str): Specifies the level.

    Example usage:

    .. code-block:: python

        >>> import logging
        >>> from gravitorch.utils.logging import disable_logging
        >>> with disable_logging('INFO'):
        ...     logging.critical('CRITICAL')
        ...     logging.info('INFO')
        ...     logging.debug('DEBUG')
        ...
        CRITICAL:root:CRITICAL
    """
    prev_level = logging.getLogger(__name__).level
    if isinstance(level, str):
        level = logging.getLevelName(level)
    logging.disable(level)
    try:
        yield
    finally:
        logging.disable(prev_level)



__all__ = ["LeNet5"]

from collections import OrderedDict

import torch
from torch import Tensor, nn

from gravitorch import constants as ct
from gravitorch.nn.utils.module_helpers import get_module_device


class LeNet5(nn.Module):
    r"""Implementation of LeNet5 network architecture.

    The main target of this network is MNIST.

    Args:
        num_classes (int): Specifies the number of classes.
        logits (bool, optional): Specifies if the output is the logits
            or the probabilities. By default, this network returns the
            logits. Default: ``True``
    """

    def __init__(
        self,
        num_classes: int,
        logits: bool = True,
        input_name: str = ct.INPUT,
        output_name: str = ct.PREDICTION,
    ):
        super().__init__()
        self.feature_extractor = nn.Sequential(
            OrderedDict(
                [
                    (
                        "conv1",
                        nn.Conv2d(in_channels=1, out_channels=6, kernel_size=(5, 5), stride=(1, 1)),
                    ),
                    ("tanh1", nn.Tanh()),
                    ("pool1", nn.AvgPool2d(kernel_size=2)),
                    (
                        "conv2",
                        nn.Conv2d(
                            in_channels=6, out_channels=16, kernel_size=(5, 5), stride=(1, 1)
                        ),
                    ),
                    ("tanh2", nn.Tanh()),
                    ("pool2", nn.AvgPool2d(kernel_size=2)),
                    (
                        "conv3",
                        nn.Conv2d(
                            in_channels=16, out_channels=120, kernel_size=(5, 5), stride=(1, 1)
                        ),
                    ),
                    ("tanh3", nn.Tanh()),
                ]
            )
        )
        self.classifier = nn.Sequential(
            OrderedDict(
                [
                    ("linear4", nn.Linear(in_features=120, out_features=84)),
                    ("tanh4", nn.Tanh()),
                    ("linear5", nn.Linear(in_features=84, out_features=num_classes)),
                ]
            )
        )
        if not logits:
            self.classifier.add_module("softmax", nn.Softmax(dim=1))

        self._input_name = str(input_name)
        self._output_name = str(output_name)

    def forward(self, tensor: Tensor) -> Tensor:
        r"""Computes the predictions of the LeNet network.

        Args:
            tensor (``torch.Tensor`` of shape
                ``(batch size, 1, 32, 32)`` and type float): Specifies
                the input images.

        Returns:
            ``torch.Tensor`` of shape ``(batch size, num classes)``
                and type float: The logits or the probabilities for
                each class.
        """
        x = self.feature_extractor(tensor)
        x = torch.flatten(x, 1)
        return self.classifier(x)

    def get_dummy_input(self, batch_size: int = 1) -> tuple[Tensor]:
        r"""Generates a dummy input for the MLP.

        Args:
            batch_size (int, optional): Specifies the batch size to
                use to generate the dummy input. Default: ``1``

        Returns:
            ``tuple[Tensor]``: A tuple with one tensor of shape
                ``(batch_size, 1, 32, 32)`` and type ``float``.
                The tensor is on the same device that this network.
        """
        device = get_module_device(self.feature_extractor)
        return (torch.randn(batch_size, 1, 32, 32, device=device),)

    def get_input_names(self) -> tuple[str]:
        r"""Gets the tuple of input names for the network.

        The order of the name should be the same that the order in
        the inputs of the forward function.

        Returns:
            tuple: The tuple of input names.
        """
        return (self._input_name,)

    def get_onnx_dynamic_axis(self) -> dict:
        r"""Gets the dynamic axes (e.g. batch size or sequence length) when
        serializing a model to ONNX format.

        See https://pytorch.org/docs/stable/onnx.html#torch.onnx.export
        to have more information on how to create the ``dict``.

        Returns:
            dict: with the dynamic axes of the input/output.
        """
        return {self._input_name: {0: "batch"}, self._output_name: {0: "batch"}}

    def get_output_names(self) -> tuple[str]:
        r"""Gets the tuple of output names for the network.

        The order of the name should be the same that the order in the
        outputs of the forward function.

        Returns:
            tuple: The tuple of output names.
        """
        return (self._output_name,)



r"""The data package contains the engine base class and some implemented
engines."""

__all__ = [
    "AlphaEngine",
    "BaseEngine",
    "EngineEvents",
]

from gravitorch.engines.alpha import AlphaEngine
from gravitorch.engines.base import BaseEngine
from gravitorch.engines.events import EngineEvents



__all__ = ["VanillaLRScheduler", "EpochLRScheduler", "IterationLRScheduler"]

from typing import Union

from gravitorch.engines.base import BaseEngine
from gravitorch.handlers.base import BaseHandler
from gravitorch.handlers.lr_monitor import EpochLRMonitor, IterationLRMonitor
from gravitorch.handlers.lr_scheduler_updater import (
    EpochLRSchedulerUpdater,
    IterationLRSchedulerUpdater,
)
from gravitorch.handlers.utils import setup_handler
from gravitorch.utils.format import str_add_indent


class VanillaLRScheduler(BaseHandler):
    r"""Implements a handler to update a learning rate (LR) scheduler and
    monitor the LR value.

    Args:
        lr_scheduler_updater (``BaseHandler`` or dict): Specifies the
            learning rate scheduler updater or its configuration. The
            LR scheduler updater is responsible to update the LR
            scheduler.
        lr_monitor (``BaseHandler`` or dict): Specifies the learning
            rate monitor or its configuration.
    """

    def __init__(
        self,
        lr_scheduler_updater: Union[BaseHandler, dict],
        lr_monitor: Union[BaseHandler, dict],
    ):
        self._lr_scheduler_updater = setup_handler(lr_scheduler_updater)
        self._lr_monitor = setup_handler(lr_monitor)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  lr_scheduler_updater={str_add_indent(str(self._lr_scheduler_updater))},\n"
            f"  lr_monitor={str_add_indent(str(self._lr_monitor))},\n"
            ")"
        )

    def attach(self, engine: BaseEngine) -> None:
        r"""Attaches the handler to update a LR scheduler and monitor the LR
        value.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        self._lr_scheduler_updater.attach(engine)
        self._lr_monitor.attach(engine)


class EpochLRScheduler(VanillaLRScheduler):
    r"""Implements a handler to update a learning rate (LR) scheduler at the end
    of each training epoch and monitor the LR value.

    This LR scheduler handler sets up:

        - an event handler to update the LR scheduler at the end of
            each training epoch
        - a LR monitor to log the learning rate value(s) at the
            beginning of each training epoch
    """

    def __init__(self):
        super().__init__(
            lr_scheduler_updater=EpochLRSchedulerUpdater(), lr_monitor=EpochLRMonitor()
        )


class IterationLRScheduler(VanillaLRScheduler):
    r"""Implements a handler to update a learning rate (LR) scheduler at the end
    of each training iteration and monitor the LR value.

    This LR scheduler handler sets up:

        - an event handler to update the LR scheduler at the end of
            each training iteration
        - a LR monitor to log the learning rate value(s) at the
            beginning of each training iteration

    Args:
        freq (int, optional): Specifies the iteration frequency used
            to monitor the learning rate. Default: ``10``
    """

    def __init__(self, freq: int = 10):
        super().__init__(
            lr_scheduler_updater=IterationLRSchedulerUpdater(),
            lr_monitor=IterationLRMonitor(freq=freq),
        )



__all__ = ["EpochLRMonitor", "IterationLRMonitor"]

import logging

from gravitorch.engines.base import BaseEngine
from gravitorch.engines.events import EngineEvents
from gravitorch.handlers.base import BaseHandler
from gravitorch.handlers.utils import add_unique_event_handler
from gravitorch.optimizers.utils import get_learning_rate_per_group
from gravitorch.utils.events import (
    ConditionalEventHandler,
    EpochPeriodicCondition,
    IterationPeriodicCondition,
)
from gravitorch.utils.exp_trackers import EpochStep, IterationStep

logger = logging.getLogger(__name__)


class EpochLRMonitor(BaseHandler):
    r"""Implements a handler to monitor the learning rate (LR) of an optimizer
    every ``freq`` epochs.

    You should use the ``EpochOptimizerMonitor`` handler if you want
    to monitor more information about the optimizer.

    Args:
        event (str, optional): Specifies the epoch-based event when
            the learning rate should be capture.
            Default: ``'train_epoch_started'``
        freq (int, optional): Specifies the epoch frequency used to
            monitor the learning rate. Default: ``1``
    """

    def __init__(self, event: str = EngineEvents.TRAIN_EPOCH_STARTED, freq: int = 1):
        self._event = str(event)
        if freq < 1:
            raise ValueError(f"freq has to be greater than 0 (received: {freq:,})")
        self._freq = int(freq)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(freq={self._freq}, event={self._event})"

    def attach(self, engine: BaseEngine) -> None:
        add_unique_event_handler(
            engine=engine,
            event=self._event,
            event_handler=ConditionalEventHandler(
                self.monitor,
                condition=EpochPeriodicCondition(engine=engine, freq=self._freq),
                handler_kwargs={"engine": engine},
            ),
        )

    def monitor(self, engine: BaseEngine) -> None:
        r"""Monitors the learning rate.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        if engine.optimizer:
            lrs = get_learning_rate_per_group(engine.optimizer)
            engine.log_metrics(
                {f"epoch/optimizer.group{i}.lr": lr for i, lr in lrs.items()},
                step=EpochStep(engine.epoch),
            )
        else:
            logger.info(
                "It is not possible to monitor the learning rate parameters because "
                "there is no optimizer"
            )


class IterationLRMonitor(BaseHandler):
    r"""Implements a handler to monitor the learning rate (LR) of an optimizer
    every ``freq`` iterations.

    You should use the ``IterationOptimizerMonitor`` handler if you
    want to monitor more information about the optimizer.

    Args:
        event (str, optional): Specifies the iteration-based event
            when the learning rate should be capture.
            Default: ``'train_iteration_started'``
        freq (int, optional): Specifies the iteration frequency used
            to monitor the learning rate. Default: ``10``
    """

    def __init__(self, event: str = EngineEvents.TRAIN_ITERATION_STARTED, freq: int = 10):
        self._event = str(event)
        if freq < 1:
            raise ValueError(f"freq has to be greater than 0 (received: {freq:,})")
        self._freq = int(freq)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(freq={self._freq}, event={self._event})"

    def attach(self, engine: BaseEngine) -> None:
        add_unique_event_handler(
            engine=engine,
            event=self._event,
            event_handler=ConditionalEventHandler(
                self.monitor,
                condition=IterationPeriodicCondition(engine=engine, freq=self._freq),
                handler_kwargs={"engine": engine},
            ),
        )

    def monitor(self, engine: BaseEngine) -> None:
        r"""Monitors the learning rate.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        if engine.optimizer:
            lrs = get_learning_rate_per_group(engine.optimizer)
            engine.log_metrics(
                {f"iteration/optimizer.group{i}.lr": lr for i, lr in lrs.items()},
                step=IterationStep(engine.iteration),
            )
        else:
            logger.info(
                "It is not possible to monitor the learning rate parameters because "
                "there is no optimizer"
            )



__all__ = ["BaseArtifact"]

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Generic, TypeVar

T = TypeVar("T")


class BaseArtifact(Generic[T], ABC):
    r"""Defines the base class to implement an artifact."""

    @abstractmethod
    def create(self, path: Path) -> None:
        r"""Creates the artifact.

        Args:
            path (``pathlib.Path``): Specifies the path where to write
                the artifact.
        """



__all__ = ["add_unique_event_handler", "setup_handler", "setup_and_attach_handlers", "to_events"]

import logging
from typing import Union

from gravitorch.engines.base import BaseEngine
from gravitorch.handlers.base import BaseHandler
from gravitorch.utils.events import BaseEventHandler
from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


def add_unique_event_handler(
    engine: BaseEngine, event: str, event_handler: BaseEventHandler
) -> None:
    r"""Adds an event handler to the engine if it was not added previously.

    This function checks if the event handler was already added to the
    engine. If not, the event handler is added to the engine otherwise
    it is not.

    Args:
        engine (``BaseEngine``): Specifies the engine used to attach
            the event handler.
        event (str): Specifies the event.
        event_handler (``BaseEventHandler``): Specifies the event
            handler.

    Example usage:

    .. code-block:: python

        # Create an engine
        >>> from gravitorch.engines import AlphaEngine
        >>> engine = AlphaEngine(...)
        # Create an event handler
        >>> def hello_handler():
        ...     print('Hello!')
        ...
        >>> from gravitorch.utils.events import VanillaEventHandler
        >>> event_handler = VanillaEventHandler(hello_handler)
        # Add an event handler to the engine
        >>> from gravitorch.handlers import add_unique_event_handler
        >>> add_unique_event_handler(engine, 'my_event', event_handler)
    """
    if engine.has_event_handler(event_handler, event):
        logger.info(f"{event_handler} is already added to the engine for '{event}' event")
    else:
        logger.info(f"Adding {event_handler} to '{event}' event")
        engine.add_event_handler(event, event_handler)


def setup_handler(handler: Union[BaseHandler, dict]) -> BaseHandler:
    r"""Sets up a handler.

    Args:
        handler (``BaseHandler`` or dict): Specifies the handler or
            its configuration.

    Returns:
        ``BaseHandler``: The handler.
    """
    if isinstance(handler, dict):
        logger.info(
            f"Initializing a handler from its configuration... {str_target_object(handler)}"
        )
        handler = BaseHandler.factory(**handler)
    return handler


def setup_and_attach_handlers(
    engine: BaseEngine,
    handlers: Union[tuple[Union[BaseHandler, dict], ...], list[Union[BaseHandler, dict]]],
) -> list[BaseHandler]:
    r"""Sets up and attaches some handlers to the engine.

    Note that if you call this function ``N`` times with the same
    handlers, the handlers will be attached ``N`` times to the engine.

    Args:
        engine (``BaseEngine``): Specifies the engine.
        handlers (list or tuple): Specifies the list of handlers or
            their configuration.

    Returns:
        list: The list of handlers attached to the engine.
    """
    list_handlers = []
    for handler in handlers:
        handler = setup_handler(handler)
        list_handlers.append(handler)
        handler.attach(engine)
    return list_handlers


def to_events(events: Union[str, tuple[str, ...], list[str]]) -> tuple[str, ...]:
    r"""Converts the input events to a tuple of events.

    If the input is a string (i.e. single event), it is converted to a
    tuple with a single event.

    Args:
        events (str or tuple or list): Specifies the input events.

    Returns:
        tuple: The tuple of events.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.handlers import to_events
        >>> to_events('my_event')
        ('my_event',)
        >>> to_events(('my_event', 'my_other_event'))
        ('my_event', 'my_other_event')
        >>> to_events(['my_event', 'my_other_event'])
        ('my_event', 'my_other_event')
    """
    if isinstance(events, str):
        return (events,)
    return tuple(events)



__all__ = ["VanillaOptimizerCreator"]

import logging
from typing import Optional

from torch.nn import Module
from torch.optim import Optimizer

from gravitorch import constants as ct
from gravitorch.creators.optimizer.base import BaseOptimizerCreator
from gravitorch.engines.base import BaseEngine
from gravitorch.optimizers import setup_optimizer

logger = logging.getLogger(__name__)


class VanillaOptimizerCreator(BaseOptimizerCreator):
    r"""Implements a vanilla optimizer creator.

    Args:
        optimizer_config (dict or ``None``, optional): Specifies the
            optimizer configuration. If ``None``, no optimizer is
            created and ``None`` will be returned by the ``create``
            method. Default: ``None``
        add_module_to_engine (bool, optional): If ``True``, the
            optimizer is added to the engine state, so the optimizer
            state is stored when the engine creates a checkpoint.
            Default: ``True``
    """

    def __init__(self, optimizer_config: Optional[dict] = None, add_module_to_engine: bool = True):
        self._optimizer_config = optimizer_config
        self._add_module_to_engine = bool(add_module_to_engine)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(add_module_to_engine={self._add_module_to_engine})"

    def create(self, engine: BaseEngine, model: Module) -> Optional[Optimizer]:
        r"""Creates an optimizer.

        This method is responsible to register the event handlers
        associated to the optimizer.

        Args:
            engine (``gravitorch.engines.BaseEngine``): Specifies an
                engine.
            model (``torch.nn.Module``): Specifies a model.

        Returns:
            ``torch.optim.Optimizer`` or ``None``: The created
                optimizer or ``None`` if there is no optimizer to
                create.
        """
        optimizer = setup_optimizer(model=model, optimizer=self._optimizer_config)
        logger.info(f"optimizer:\n{optimizer}")
        if self._add_module_to_engine and optimizer is not None:
            logger.info(f"Adding an optimizer to the engine (key: {ct.OPTIMIZER})...")
            engine.add_module(ct.OPTIMIZER, optimizer)
        return optimizer



__all__ = [
    "TensorDictShufflerIterDataPipe",
    "shuffle_tensors",
    "shuffle_tensor_mapping",
    "get_first_dimension",
]

import logging
from collections.abc import Iterator, Mapping, Sequence
from typing import Optional, Union

import numpy as np
import torch
from torch import Tensor
from torch.utils.data import IterDataPipe

from gravitorch.utils.format import str_add_indent
from gravitorch.utils.seed import get_torch_generator
from gravitorch.utils.tensor import permute

logger = logging.getLogger(__name__)


class TensorDictShufflerIterDataPipe(IterDataPipe[dict]):
    r"""Implements a source DataPipe to shuffle ``torch.Tensor``s in
    dictionaries.

    For each dictionary, the ``torch.Tensor``s are shuffled with the
    same permutation on a given dimension. This function shuffles
    each tensor on one dimension. It is possible to control the
    dimension used to shuffle the tensors with the argument ``dim``.
    There are two approaches:

        - If ``dim`` is an integer, all the tensors are shuffled on the
            dimension indicated by ``dim``. All the tensors should have
            the same dimension for the specified dimension
            ``(..., dimension, *)`` where ``*`` means any number of
            dimensions.
        - If ``dim`` is a dictionary, each tensor can be shuffled on a
            specific dimension. ``dim`` should indicate the dimension
            to use for each tensor. The keys in this dictionary should
            include keys that are in the input ``mapping``. Note that
            ``dim`` can contain only a subset of keys in the input
            ``mapping``. See example in ``shuffle_tensor_mapping``.

    Args:
        datapipe (``IterDataPipe``): Specifies the source
            iterable DataPipe.
        dim (int or dict, optional): Specifies the dimension used to
            shuffle the mapping. Default: ``0``
        random_seed (int, optional): Specifies the random seed used
            to shuffle the ``torch.Tensor``s.
            Default: ``3510637111256283951``

    Example usage:

    .. code-block:: python

        >>> from torch.utils.data import IterDataPipe
        >>> from gravitorch.data.datapipes.iter import TensorDictShuffler
        >>> class MyIterDataPipe(IterDataPipe[dict]):
        ...     def __iter__(self) -> Iterator[dict]:
        ...         for i in range(3):
        ...             yield {"key": torch.arange(4) + i}
        ...
        >>> list(TensorDictShuffler(MyIterDataPipe()))
        [{'key': tensor([2, 3, 0, 1])},
         {'key': tensor([3, 2, 4, 1])},
         {'key': tensor([2, 5, 4, 3])}]
        >>> list(TensorDictShuffler(MyIterDataPipe(), dim={'key': 0}))
        [{'key': tensor([2, 3, 0, 1])},
         {'key': tensor([3, 2, 4, 1])},
         {'key': tensor([2, 5, 4, 3])}]
    """

    def __init__(
        self,
        datapipe: IterDataPipe[Mapping],
        dim: Union[int, dict] = 0,
        random_seed: int = 3510637111256283951,
    ):
        self._datapipe = datapipe
        self._dim = dim
        self._generator = get_torch_generator(random_seed)

    def __iter__(self) -> Iterator[dict]:
        for data in self._datapipe:
            yield shuffle_tensor_mapping(data, dim=self._dim, generator=self._generator)

    def __len__(self) -> int:
        return len(self._datapipe)

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  dim={self._dim},\n"
            f"  random_seed={self.random_seed},\n"
            f"  datapipe={str_add_indent(self._datapipe)},\n)"
        )

    @property
    def random_seed(self) -> int:
        r"""int: The random seed used to initialize the pseudo random generator."""
        return self._generator.initial_seed()


def shuffle_tensors(
    tensors: Sequence[Tensor],
    dim: int = 0,
    generator: Optional[torch.Generator] = None,
) -> list[Tensor]:
    r"""Shuffles tensors with the same permutation on a given dimension.

    This function shuffles the tensors only on a dimension. It is
    possible to control the dimension used to shuffle the tensors with
    the argument ``dim``. All the tensors should have the same
    dimension for the specified dimension ``(..., dimension, *)`` where
    ``*`` means any number of dimensions.

    Args:
        tensors (``Sequence``): Specifies the tensors to shuffle.
        dim (int, optional): Specifies the dimension used to shuffle
            the tensors. Default: ``0``
        generator (``torch.Generator`` or ``None``, optional):
            Specifies an optional random generator. Default: ``None``

    Returns:
        ``list``: The shuffled tensors.

    Raises:
        ValueError if the tensors do not have the same shape for the
            common dimension.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.data.datapipes.iter.shuffling import shuffle_tensors
        >>> shuffle_tensors([torch.arange(4), torch.arange(20).view(4, 5)])
        [
            torch.tensor([0, 2, 1, 3]),
            torch.tensor([[0, 1, 2, 3, 4],
                          [10, 11, 12, 13, 14],
                          [5, 6, 7, 8, 9],
                          [15, 16, 17, 18, 19]]),
        ]
        >>> shuffle_tensors([torch.arange(4).view(1, 4), torch.arange(20).view(5, 4)], dim=1)
        [
            torch.tensor([[0, 2, 1, 3]]),
            torch.tensor([[0, 2, 1, 3],
                          [4, 6, 5, 7],
                          [8, 10, 9, 11],
                          [12, 14, 13, 15],
                          [16, 18, 17, 19]]),
        ]
    """
    dims = [tensor.shape[dim] for tensor in tensors]
    if len(set(dims)) > 1:
        raise ValueError(
            f"The tensors do not have the same shape for the common dimension (dim={dim}): {dims}"
        )
    permutation = torch.randperm(tensors[0].shape[dim], generator=generator)
    return [permute(tensor, permutation, dim) for tensor in tensors]


def shuffle_tensor_mapping(
    mapping: Mapping, dim: Union[int, dict] = 0, generator: Optional[torch.Generator] = None
) -> dict:
    r"""Shuffles the tensors that are in a mapping with the same permutation.

    This function shuffles each tensor on one dimension. It is
    possible to control the dimension used to shuffle the tensors
    with the argument ``dim``. There are two approaches:

        - If ``dim`` is an integer, all the tensors are shuffled on the
            dimension indicated by ``dim``. All the tensors should have
            the same dimension for the specified dimension
            ``(..., dimension, *)`` where ``*`` means any number of
            dimensions.
        - If ``dim`` is a dictionary, each tensor can be shuffled on a
            specific dimension. ``dim`` should indicate the dimension
            to use for each tensor. The keys in this dictionary should
            include keys that are in the input ``mapping``. Note that
            ``dim`` can contain only a subset of keys in the input
            ``mapping``. See example below.

    Args:
        mapping (``Mapping``): Specifies the mapping with the tensors
            to shuffle.
        dim (int or dict, optional): Specifies the dimension used to
            shuffle the mapping. Default: ``0``
        generator (``torch.Generator`` or ``None``, optional):
            Specifies an optional random generator. Default: ``None``

    Returns:
        ``dict``: A dictionary with the shuffled tensors.

    Raises:
        ValueError if the tensors do not have the same shape for the
            common dimension.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.data.datapipes.iter.shuffling import shuffle_tensor_mapping
        # Shuffle all the tensors on dimension 0 (default)
        >>> shuffle_tensor_mapping({"key1": torch.arange(4), "key2": torch.arange(20).view(4, 5)})
        {
            "key1": tensor([0, 2, 1, 3]),
            "key2": tensor([[0, 1, 2, 3, 4],
                            [10, 11, 12, 13, 14],
                            [5, 6, 7, 8, 9],
                            [15, 16, 17, 18, 19]]),
        }
        # Shuffle all the tensors on dimension 1
        >>> shuffle_tensor_mapping(
        ...     {"key1": torch.arange(4).view(1, 4), "key2": torch.arange(20).view(5, 4)},
        ...     dim=1,
        ... )
        {
            "key1": tensor([[0, 2, 1, 3]]),
            "key2": tensor([[0, 2, 1, 3],
                            [4, 6, 5, 7],
                            [8, 10, 9, 11],
                            [12, 14, 13, 15],
                            [16, 18, 17, 19]]),
        }
        # Shuffle the tensor "key1" on dimension 0 and the tensor "key2" on dimension 1
        >>> shuffle_tensor_mapping(
        ...     {"key1": torch.arange(4), "key2": torch.arange(20).view(5, 4)},
        ...     dim={'key1': 0, 'key2': 1}
        ... )
        {
            'key1': tensor([1, 2, 0, 3]),
            'key2': tensor([[1,  2,  0,  3],
                            [5,  6,  4,  7],
                            [9, 10,  8, 11],
                            [13, 14, 12, 15],
                            [17, 18, 16, 19]]),
        }
        # Shuffle only the tensor "key2" on dimension 1
        >>> shuffle_tensor_mapping(
        ...     {"key1": torch.arange(4), "key2": torch.arange(20).view(5, 4)},
        ...     dim={'key2': 1},
        ... )
        {
            'key1': tensor([0, 1, 2, 3]),
            'key2': tensor([[3,  1,  0,  2],
                            [7,  5,  4,  6],
                            [11,  9,  8, 10],
                            [15, 13, 12, 14],
                            [19, 17, 16, 18]]),
        }
    """
    dims = dim if isinstance(dim, dict) else {key: dim for key in mapping.keys()}
    if len(dims) == 0:  # No tensor to permute
        return {key: tensor for key, tensor in mapping.items()}
        # Check if the common dimensions are the same
    valid_dims = [mapping[key].shape[dims[key]] for key in mapping.keys() if key in dims]
    if len(set(valid_dims)) > 1:
        raise ValueError(
            f"The tensors do not have the same shape for the common dimension: {valid_dims}"
        )
    # Permute the tensors
    key = next(iter(dims.keys()))
    permutation = torch.randperm(mapping[key].shape[dims[key]], generator=generator)
    output = {}
    for key, tensor in mapping.items():
        if key in dims:
            tensor = permute(tensor, permutation, dims[key])
        output[key] = tensor
    return output


def get_first_dimension(data: Union[Tensor, np.ndarray, list, tuple]) -> int:
    r"""Finds the first dimension value.

    This function finds the first dimension value by using the
    following rules:

        - if the input is a ``torch.Tensor`` or ``numpy.ndarray``,
            the first dimension value is returned.
        - if the input is a list or tuple, the length is returned.

    Args:
        data: Specifies the data to get the first dimension value.

    Returns:
        int: The first dimension value.

    Raises:
        TypeError: if the input type is not supported.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.data.datapipes.iter.shuffling import get_first_dimension
        # torch tensor
        >>> import torch
        >>> get_first_dimension(torch.ones(5, 4))
        5
        # NumPy ndarray
        >>> import numpy
        >>> get_first_dimension(numpy.ones((7, 4)))
        7
        # List/tuple
        >>> get_first_dimension([1, 3, 4])
        3
        >>> get_first_dimension((5, 2, 1, 4))
        4
    """
    if torch.is_tensor(data) or isinstance(data, np.ndarray):
        return data.shape[0]
    if isinstance(data, (list, tuple)):
        return len(data)
    raise TypeError(
        f"Incorrect type. The supported types are: torch.Tensor, numpy.ndarray, list and tuple. "
        f"(received: {data})"
    )



__all__ = ["ConsolidateOptimizerState"]

import logging

from gravitorch import distributed as dist
from gravitorch.engines.base import BaseEngine
from gravitorch.engines.events import EngineEvents
from gravitorch.handlers.base import BaseHandler
from gravitorch.handlers.utils import add_unique_event_handler
from gravitorch.utils.events import VanillaEventHandler

logger = logging.getLogger(__name__)


class ConsolidateOptimizerState(BaseHandler):
    r"""Implements a handler to consolidate the state dict of an optimizer.

    Args:
        event (str, optional): Specifies the event when the optimizer
            state dict is consolidated.
            Default: ``'train_epoch_completed'``
        recipient_rank (int, optional): Specifies on which rank to
            materialize the full state dict. ``-1`` is a special
            value, which means that all ranks should have the state.
            Default: ``0``
    """

    def __init__(self, event: str = EngineEvents.TRAIN_EPOCH_COMPLETED, recipient_rank: int = 0):
        self._event = str(event)
        self._recipient_rank = int(recipient_rank)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(event={self._event}, "
            f"recipient_rank={self._recipient_rank})"
        )

    def attach(self, engine: BaseEngine) -> None:
        r"""Attaches the event handler to the engine.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        add_unique_event_handler(
            engine=engine,
            event=self._event,
            event_handler=VanillaEventHandler(
                self.consolidate,
                handler_kwargs={"engine": engine},
            ),
        )

    def consolidate(self, engine: BaseEngine) -> None:
        r"""Consolidate the optimizer state dict.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        if not engine.optimizer:
            logger.info(
                "It is not possible to consolidate the optimizer state dict "
                "because there is no optimizer"
            )
        elif hasattr(engine.optimizer, "consolidate_state_dict"):
            logger.info(f"Consolidating the optimizer state dict on rank {self._recipient_rank}...")
            engine.optimizer.consolidate_state_dict(self._recipient_rank)
            dist.barrier()
        else:
            logger.info("The optimizer does not have a 'consolidate_state_dict' method")



from __future__ import annotations

__all__ = ["AlphaEngine"]

import logging
from typing import TYPE_CHECKING, Any

from torch.nn import Module
from torch.optim import Optimizer

from gravitorch import constants as ct
from gravitorch.creators.core.base import BaseCoreCreator, setup_core_creator
from gravitorch.distributed import comm as dist
from gravitorch.engines.base import BaseEngine
from gravitorch.engines.events import EngineEvents
from gravitorch.lr_schedulers import LRSchedulerType
from gravitorch.utils.artifacts import BaseArtifact
from gravitorch.utils.engine_states import BaseEngineState, setup_engine_state
from gravitorch.utils.evaluation_loops import BaseEvaluationLoop, setup_evaluation_loop
from gravitorch.utils.events import BaseEventHandler, EventManager
from gravitorch.utils.exp_trackers import BaseExpTracker, Step, setup_exp_tracker
from gravitorch.utils.format import str_add_indent
from gravitorch.utils.history import BaseHistory
from gravitorch.utils.timing import timeblock
from gravitorch.utils.training_loops import BaseTrainingLoop, setup_training_loop

if TYPE_CHECKING:
    from gravitorch.datasources import BaseDataSource

logger = logging.getLogger(__name__)


class AlphaEngine(BaseEngine):
    r"""Implements an engine to train or evaluate a model.

    The core engine modules of this engine are created by a core
    engine modules creator.

    Args:
        core_creator (``BaseCoreCreator`` or dict):
            Specifies the core engine modules creators or its
            configuration. This object creates the data module, model,
            optimizer and LR scheduler.
        state (``BaseEngineState`` or dict or ``None``): Specifies
            the engine state or its configuration. If ``None``, a
            ``VanillaEngineState`` object is created. The state is
            used to represent all the modules added to the engine,
            but the event system is not include in the engine state.
            Default: ``None``
        exp_tracker (``BaseExpTracker`` or dict or ``None``):
            Specifies the experiment tracker or its configuration.
            If ``None``, a ``NoOpExpTracker`` object is created.
            The experiment tracker is used to track some artifacts
            in a third-party library.
            Default: ``None``
        training_loop (``BaseTrainingLoop`` or dict or ``None``):
            Specifies the training loop or its configuration.
            If ``None``, a ``VanillaTrainingLoop`` object is created.
            Default: ``None``
        evaluation_loop (``BaseEvaluationLoop`` or dict or ``None``):
            Specifies the evaluation loop or its configuration.
            If ``None``, a ``VanillaEvaluationLoop`` object is
            created. Default: ``None``
    """

    def __init__(
        self,
        core_creator: BaseCoreCreator | dict,
        state: BaseEngineState | dict | None = None,
        exp_tracker: BaseExpTracker | dict | None = None,
        training_loop: BaseTrainingLoop | dict | None = None,
        evaluation_loop: BaseEvaluationLoop | dict | None = None,
    ):
        self._event_manager = EventManager()
        self._state = self._setup_state(state)
        self._exp_tracker = self._setup_exp_tracker(exp_tracker)
        self._training_loop = self._setup_training_loop(training_loop)
        self._evaluation_loop = self._setup_evaluation_loop(evaluation_loop)

        core_creator = setup_core_creator(core_creator)
        logger.info(f"core_creator:\n{core_creator}")
        (
            self._data_source,
            self._model,
            self._optimizer,
            self._lr_scheduler,
        ) = core_creator.create(self)
        # TODO: add the modules if there are not added yet.

        self._should_terminate = False

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  data_source={str_add_indent(self._data_source)},\n"
            f"  model={str_add_indent(self._model)},\n"
            f"  optimizer={str_add_indent(self._optimizer)},\n"
            f"  lr_scheduler={str_add_indent(self._lr_scheduler)},\n"
            f"  training_loop={str_add_indent(self._training_loop)},\n"
            f"  evaluation_loop={str_add_indent(self._evaluation_loop)},\n"
            f"  state={str_add_indent(self._state)},\n"
            f"  event_manager={str_add_indent(self._event_manager)},\n"
            f"  exp_tracker={str_add_indent(self._exp_tracker)},\n"
            ")"
        )

    @property
    def data_source(self) -> BaseDataSource:
        return self._data_source

    @property
    def epoch(self) -> int:
        return self._state.epoch

    @property
    def iteration(self) -> int:
        return self._state.iteration

    @property
    def lr_scheduler(self) -> LRSchedulerType | None:
        return self._lr_scheduler

    @property
    def max_epochs(self) -> int:
        return self._state.max_epochs

    @property
    def model(self) -> Module:
        return self._model

    @property
    def optimizer(self) -> Optimizer | None:
        return self._optimizer

    @property
    def random_seed(self) -> int:
        return self._state.random_seed

    @property
    def should_terminate(self) -> bool:
        return self._should_terminate

    def add_event_handler(self, event: str, event_handler: BaseEventHandler) -> None:
        self._event_manager.add_event_handler(event, event_handler)

    def add_history(self, history: BaseHistory, key: str | None = None) -> None:
        self._state.add_history(history=history, key=key)

    def add_module(self, name: str, module: Any) -> None:
        self._state.add_module(name=name, module=module)

    def create_artifact(self, artifact: BaseArtifact) -> None:
        self._exp_tracker.create_artifact(artifact)

    def eval(self) -> None:
        with timeblock("=== Evaluation time: {time} ==="):
            logger.info("Launching evaluation procedures")
            dist.barrier()
            self.fire_event(EngineEvents.STARTED)
            self._evaluation_loop.eval(self)
            dist.barrier()
            self.fire_event(EngineEvents.COMPLETED)
            self._exp_tracker.flush()
            logger.info("Ending evaluation procedures")

    def fire_event(self, event: str) -> None:
        self._event_manager.fire_event(event)

    def get_history(self, key: str) -> BaseHistory:
        return self._state.get_history(key)

    def get_histories(self) -> dict[str, BaseHistory]:
        return self._state.get_histories()

    def get_module(self, name: str) -> Any:
        return self._state.get_module(name)

    def has_event_handler(self, event_handler: BaseEventHandler, event: str | None = None) -> bool:
        return self._event_manager.has_event_handler(event_handler, event)

    def has_history(self, key: str) -> bool:
        return self._state.has_history(key)

    def has_module(self, name: str) -> bool:
        return self._state.has_module(name)

    def increment_epoch(self, increment: int = 1) -> None:
        self._state.increment_epoch(increment)

    def increment_iteration(self, increment: int = 1) -> None:
        self._state.increment_iteration(increment)

    def load_state_dict(self, state_dict: dict) -> None:
        self._state.load_state_dict(state_dict)

    def log_figure(
        self,
        key: str,
        figure: matplotlib.pyplot.Figure,  # noqa: F821
        step: Step | None = None,
    ) -> None:
        self._exp_tracker.log_figure(key, figure, step)

    def log_figures(
        self,
        figures: dict[str, matplotlib.pyplot.Figure],  # noqa: F821
        step: Step | None = None,
    ) -> None:
        self._exp_tracker.log_figures(figures, step)

    def log_metric(self, key: str, value: int | float, step: Step | None = None):
        self._state.get_history(key).add_value(value, step.step if step else step)
        self._exp_tracker.log_metric(key, value, step)

    def log_metrics(self, metrics: dict[str, int | float], step: Step | None = None) -> None:
        for key, value in metrics.items():
            self._state.get_history(key).add_value(value, step.step if step else step)
        self._exp_tracker.log_metrics(metrics, step)

    def remove_event_handler(self, event: str, event_handler: BaseEventHandler) -> None:
        self._event_manager.remove_event_handler(event, event_handler)

    def remove_module(self, name: str) -> None:
        self._state.remove_module(name)

    def state_dict(self) -> dict[str, Any]:
        return self._state.state_dict()

    def terminate(self) -> None:
        self._should_terminate = True

    def train(self) -> None:
        with timeblock("=== Training time: {time} ==="):
            logger.info("Launching training procedures")
            self.fire_event(EngineEvents.STARTED)
            self.fire_event(EngineEvents.TRAIN_STARTED)
            if not self._should_terminate:
                for _ in range(self.epoch + 1, self.max_epochs):
                    self.increment_epoch()
                    dist.barrier()
                    self.fire_event(EngineEvents.EPOCH_STARTED)

                    self._training_loop.train(self)
                    self._evaluation_loop.eval(self)

                    dist.barrier()
                    self.fire_event(EngineEvents.EPOCH_COMPLETED)
                    self._log_best_metrics()
                    self._exp_tracker.flush()

                    if self._should_terminate:
                        break

                dist.barrier()
                self.fire_event(EngineEvents.TRAIN_COMPLETED)

            self.fire_event(EngineEvents.COMPLETED)
            dist.barrier()
            self._exp_tracker.flush()
            logger.info("Ending training procedures...")

    def _log_best_metrics(self) -> None:
        r"""Logs the best value of each scalar comparable metric."""
        dist.barrier()
        best_metrics = self._state.get_best_values()
        best_metrics = {
            key: value for key, value in best_metrics.items() if isinstance(value, (int, float))
        }
        self._exp_tracker.log_best_metrics(best_metrics)

    def _setup_evaluation_loop(
        self, evaluation_loop: BaseEvaluationLoop | dict | None
    ) -> BaseEvaluationLoop:
        r"""Sets up the evaluation loop.

        The evaluation loop is instantiated from its configuration
        by using the ``BaseEvaluationLoop`` factory function.

        Args:
            evaluation_loop (``BaseEvaluationLoop`` or dict or None):
                Specifies the evaluation loop or its configuration.
                If ``None``, the ``VanillaEvaluationLoop`` is
                instantiated.

        Returns:
            ``BaseEvaluationLoop``: The evaluation loop.
        """
        evaluation_loop = setup_evaluation_loop(evaluation_loop)
        logger.info(f"evaluation loop:\n{evaluation_loop}")
        self.add_module(ct.EVALUATION_LOOP, evaluation_loop)
        return evaluation_loop

    def _setup_exp_tracker(self, exp_tracker: BaseExpTracker | dict | None) -> BaseExpTracker:
        r"""Sets up the experiment tracker.

        If the input is None, the no-operation experiment tracker is
        used. The experiment tracker is instantiated from its
        configuration by using the ``BaseExpTracker`` factory function.

        Args:
            exp_tracker (``BaseExpTracker`` or dict or None):
                Specifies the experiment tracker or its configuration.

        Returns:
            ``BaseExpTracker``: The (instantiated) experiment tracker.
        """
        exp_tracker = setup_exp_tracker(exp_tracker)
        logger.info(f"experiment tracker:\n{exp_tracker}")
        if not exp_tracker.is_activated():
            exp_tracker.start()
        return exp_tracker

    def _setup_state(self, state: BaseEngineState | dict | None) -> BaseEngineState:
        r"""Sets up the engine state.

        The engine state is instantiated from its configuration by
        using the ``BaseEngineState`` factory function.

        Args:
            state (``BaseEngineState`` or dict or None):
                Specifies the engine state or its configuration.
                If ``None``, the ``VanillaEngineState`` is
                instantiated.

        Returns:
            ``BaseEngineState``: The engine state.
        """
        state = setup_engine_state(state)
        logger.info(f"state:\n{state}")
        return state

    def _setup_training_loop(
        self, training_loop: BaseTrainingLoop | dict | None
    ) -> BaseTrainingLoop:
        r"""Sets up the training loop.

        The training loop is instantiated from its configuration by
        using the ``BaseTrainingLoop`` factory function.

        Args:
            training_loop (``BaseTrainingLoop`` or dict or None):
                Specifies the training loop or its configuration.
                If ``None``, the ``VanillaTrainingLoop`` is
                instantiated.

        Returns:
            ``BaseTrainingLoop``: The training loop.
        """
        training_loop = setup_training_loop(training_loop)
        logger.info(f"training loop:\n{training_loop}")
        self.add_module(ct.TRAINING_LOOP, training_loop)
        return training_loop



r"""This module defines a criterion wrapper to make PyTorch criteria compatible
with ``gravitorch.models.VanillaModel``."""

__all__ = ["VanillaLoss"]

from typing import Union

from torch import Tensor
from torch.nn import Module

from gravitorch import constants as ct
from gravitorch.nn import setup_nn_module


class VanillaLoss(Module):
    r"""Implements a wrapper to make compatible most of the PyTorch loss
    functions (a.k.a. criterion) with ``gravitorch.models.VanillaModel``.

    This wrapper assumes the loss function has two inputs:

        - a tensor of prediction.
        - a tensor of target

    The shape and type of the tensors depend on the loss function used.

    Args:
        criterion (``torch.nn.Module`` or dict): Specifies the loss
            function (a.k.a. criterion) or its configuration.
        prediction_key (str): Specifies the prediction key.
            Default: ``"prediction"``.
        target_key (str): Specifies the target key.
            Default: ``"target"``.

    Example usage:

    .. code-block:: python

        >>> from torch import nn
        >>> from gravitorch.models.criterions import VanillaLoss
        # Initialization with a nn.Module
        >>> criterion = VanillaLoss(criterion=nn.MSELoss())
        # Initialization with a config
        >>> criterion = VanillaLoss(criterion={'_target_': 'torch.nn.MSELoss'})
        # Customize keys.
        >>> criterion = VanillaLoss(
        ...     criterion=nn.MSELoss(),
        ...     prediction_key='my_prediction',
        ...     target_key='my_target',
        ... )
        >>> net_out = {'my_prediction': ...}
        >>> batch = {'my_target': ...}
        >>> loss = criterion(net_out, batch)
        {'loss': torch.tensor(...)}
    """

    def __init__(
        self,
        criterion: Union[Module, dict],
        prediction_key: str = ct.PREDICTION,
        target_key: str = ct.TARGET,
    ):
        super().__init__()
        self.criterion = setup_nn_module(criterion)
        self._prediction_key = prediction_key
        self._target_key = target_key

    def forward(self, net_out: dict, batch: dict) -> dict[str, Tensor]:
        r"""Computes the loss value given the network output and the batch.

        Args:
            net_out (dict): Specifies the network output
                which contains the prediction.
            batch (dict): Specifies the batch which contains the target.

        Returns:
            dict: a dict with the loss value.
        """
        prediction = self._get_prediction_from_net_out(net_out)
        target = self._get_target_from_batch(batch)
        return {ct.LOSS: self.criterion(prediction, target)}

    def _get_prediction_from_net_out(self, net_out: dict) -> Tensor:
        r"""Gets the prediction from the network output.

        Args:
            net_out (dict): Specifies the network output
                which contains the prediction.

        Returns:
            ``torch.Tensor``: The prediction
        """
        return net_out[self._prediction_key]

    def _get_target_from_batch(self, batch: dict) -> Tensor:
        r"""Gets the target from the batch. The target is the tensor with the
        key 'target'.

        Args:
            batch (dict): Specifies the batch which contains the target.

        Returns:
            ``torch.Tensor``: The target
        """
        return batch[self._target_key]
